import numpy as np
import pandas as pd


def merge_consortia(data):
    """
    :param data: pandas.DataFrame. The intermediate astrometric data for a Hipparcos 1 source.
    :return: merged_data: pandas.DataFrame.
    The input data with merged residuals, errors, sin(scan_angle), cos(scan_angle), epochs etc. The
    residuals and errors (along-scan errors) are merged according the Hipparcos and Tycho Catalogues Vol 3
    Section "Astrometric Catalogue Merging" (page 377).

    This function takes the merged sin of the scan angle and epochs etc to be the weighted mean of
    those from both consortia (weighted by the covariance between observations, see `More' below).

    Before merging, observations that were rejected by the Hipparcos team (flagged with lowercase f or n in
    the consortia (IA2) column) are removed and not used for merging. See the description of Field IA2 in
    the Hipparcos and Tycho Catalogues Vol 1, page 260.

    More: The intermediate astrometric data provides the correlation coefficients between the values
    for the residuals and errors obtained by the NDAC and FAST consortia. One can then rebuild the covariance
    between the measurements of those two consortia (i.e. treating their results as correlated measurements of
    some unknown value -- which is the merged value we are after). Given the covariance matrix for one orbit, C
    from equation 17.11 of the Hipparcos and Tycho Catalogues Vol 3 (page 377):
            C = np.asarray([[errF[i]**2, errN[i]*errF[i]*corr[i]],
                        [errN[i]*errF[i]*corr[i], errN[i]**2]])
    Where errN[i] is the NDAC along-scan error at the ith epoch, and errF is the FAST error, and corr[i] is the
    correlation between the two measurements (column IA10). The merged residual is then the value which minimizes
    the chisquared = (residual - merged_residual) C^(-1) (residual - merged_residual)
    with matrix dot products implied and residual = the vector of residuals = [NDAC residual, FAST residual]
    The merged error is 1/sum(C^(-1))**0.5
    Where C^(-1) is the matrix inverse of the covariance matrix C.
    """
    # exclude observations that were rejected for the merged solution (those with n, f instead of N, F)
    data.drop(np.argwhere(np.logical_or((data['IA2'] == 'n').to_numpy(),
                                        (data['IA2'] == 'f').to_numpy())).flatten(), inplace=True)
    # We transform to Numpy arrays because accessing and editing panda arrays is slower by factors of tens.
    cols_to_merge = ['A1', 'IA3', 'IA4', 'IA5', 'IA6', 'IA7', 'IA8', 'IA9', 'IA10']
    data_asarray = data[cols_to_merge].to_numpy()
    # merge data
    merged_data = _merge_orbits(data_asarray)
    merged_data = pd.DataFrame(merged_data, columns=cols_to_merge)
    merged_data['IA2'] = 'M'
    return merged_data


def _merge_orbits(data):
    """
    :param data: ndarray. Intermediate Data for a single Hipparcos 1 orbit.
                data must be a 2D array such that
                data[:, -1] = column IA10  (correlation coeff). Note data[:, -1] is the last column of data
                data[:, -2] = column IA9  (formal error along scan)
                data[:, -3] = column IA8  (residual along scan)
                data[:, 0] = column A1  (the orbit number)
    :return merged_orbit: ndarray. Merged intermediate astrometric data for a single orbit.
    Will have shape (m, N), where m is the number of unique orbits in data, i.e. m = len(np.unique(data[:, 0]))
    """
    # find single consortia orbits (either NDAC or FAST but not both)
    orbs, indices, counts = np.unique(data[:, 0], return_index=True, return_counts=True)
    # omit merging orbits with only a single consortium
    single_orbits = data[indices[counts == 1]]
    data = np.delete(data, indices[counts == 1], axis=0)
    # generate weights for merging
    err1, err2, corr = data[::2, -2], data[1::2, -2], data[::2, -1]
    icov = np.linalg.pinv(np.array([[err1 ** 2, err2 * err1 * corr],
                                    [err2 * err1 * corr, err2 ** 2]]).T.reshape((-1, 2, 2)))
    # constructing the weights via the best linear estimator method (which seeks to minimize the variance).
    weights = np.array([np.sum(np.dot([1, 0], icov), axis=1), np.sum(np.dot([0, 1], icov), axis=1)])
    weights /= np.sum(np.sum(icov, axis=-1), axis=-1)  # normalize weights by sum of elements of each icov matrix.
    # merge data
    merged_data = weights[0].reshape(-1, 1) * data[::2] + weights[1].reshape(-1, 1) * data[1::2]
    # evaluate variances on the BLUE estimator x'. Var(x') = w0^2*Var(x0) + w1^2*Var(x1) + 2 * w1 * w0 * Cov(x1, x2)
    # where w1, w2 are the weights and x' is the merged result computed above.
    merged_data[:, -2] = weights[0] ** 2 * err1 ** 2 + weights[1] ** 2 * err2 ** 2 + \
                         2 * corr * err1 * err2 * weights[0] * weights[1]
    # convert the variances to standard deviations
    merged_data[:, -2] = np.sqrt(merged_data[:, -2])
    # stack the merged_data back with the orbits with a single consortia (that could therefore not be merged)
    merged_data = np.vstack((merged_data, single_orbits))
    return merged_data[np.argsort(merged_data[:, 0])]


def safe_concatenate(a, b):
    if a is None and b is None:
        return None
    if a is None and b is not None:
        return b
    if a is not None and b is None:
        return a
    return np.concatenate([a, b])



"""
Utility functions for the validation scripts.
"""
from htof.parse import DataParser, HipparcosOriginalData, HipparcosRereductionDVDBook, HipparcosRereductionJavaTool
from htof.fit import AstrometricFitter
from htof.sky_path import parallactic_motion, earth_ephemeris
from astropy import time
from astropy.coordinates import Angle
from astropy.table import Table
from glob import glob
import os
import warnings
import numpy as np


def refit_hip_fromdata(data: DataParser, fit_degree, cntr_RA=Angle(0, unit='degree'), cntr_Dec=Angle(0, unit='degree'),
                       use_parallax=False):
    data.calculate_inverse_covariance_matrices()
    # generate parallax motion
    jyear_epoch = time.Time(data.julian_day_epoch(), format='jd', scale='tcb').jyear
    # note that ra_motion and dec_motion are in degrees here.
    # generate sky path
    year_epochs = jyear_epoch - time.Time(1991.25, format='decimalyear', scale='tcb').jyear
    ra_motion, dec_motion = parallactic_motion(jyear_epoch, cntr_RA.degree, cntr_Dec.degree, 'degree',
                                               time.Time(1991.25, format='decimalyear', scale='tcb').jyear,
                                               ephemeris=earth_ephemeris)  # Hipparcos was in a geostationary orbit.

    ra_resid = Angle(data.residuals.values * np.sin(data.scan_angle.values), unit='mas')
    dec_resid = Angle(data.residuals.values * np.cos(data.scan_angle.values), unit='mas')
    # instantiate fitter
    fitter = AstrometricFitter(data.inverse_covariance_matrix, year_epochs, normed=False,
                               use_parallax=use_parallax, fit_degree=fit_degree,
                               parallactic_pertubations={'ra_plx': Angle(ra_motion, 'degree').mas,
                                                         'dec_plx': Angle(dec_motion, 'degree').mas})

    fit_coeffs, errors, chisq = fitter.fit_line(ra_resid.mas, dec_resid.mas, return_all=True)
    if not use_parallax:
        fit_coeffs = np.hstack([[0], fit_coeffs])
        errors = np.hstack([[0], errors])
    # pad so coeffs and errors are 9 long.
    errors = np.pad(errors, (0, 9 - len(fit_coeffs)))
    fit_coeffs = np.pad(fit_coeffs, (0, 9 - len(fit_coeffs)))
    # calculate the chisquared partials
    sin_scan = np.sin(data.scan_angle.values)
    cos_scan = np.cos(data.scan_angle.values)
    dt = data.epoch - 1991.25
    chi2_vector = (2 * data.residuals.values / data.along_scan_errs.values ** 2 * np.array(
        [sin_scan, cos_scan, dt * sin_scan, dt * cos_scan])).T
    chi2_partials = np.sum(chi2_vector, axis=0) ** 2
    return fit_coeffs, errors, chisq, chi2_partials


def refit_hip1_object(iad_dir, hip_id, hip_dm_g=None, use_parallax=False):
    data = HipparcosOriginalData()
    data.parse(star_id=hip_id, intermediate_data_directory=iad_dir)
    fname = os.path.join(iad_dir, hip_id + '.txt')

    plx, cntr_RA, cntr_Dec, pmRA, pmDec, soltype = get_cat_values_hip1(fname)
    soltype = soltype.strip()
    accRA, accDec, jerkRA, jerkDec = 0, 0, 0, 0
    fit_degree = {'5': 1}.get(soltype, None)
    if hip_dm_g is not None and (soltype == '7' or soltype == '9'):
        # do the fit for seven/nine parameter fits if we have the 7th and 9th parameters.
        idx = np.searchsorted(hip_dm_g['hip_id'].data, int(hip_id))  # int(hip_id) strips leading zeroes.
        accRA, accDec, jerkRA, jerkDec = hip_dm_g[idx][['accRA', 'accDec', 'jerkRA', 'jerkDec']]
        fit_degree = {'5': 1, '7': 2, '9': 3}.get(soltype, None)

    if fit_degree is not None:
        # abscissa residuals are always with respect to 5p solution for hip1. Do not feed accRA, etc..
        # when reconstructing the skypath
        diffs, errors, chisq, chi2_partials = refit_hip_fromdata(data, fit_degree, cntr_RA=cntr_RA, cntr_Dec=cntr_Dec,
                                                                 use_parallax=use_parallax)
        if soltype == '7' or soltype == '9':
            # to account for the 0.81 yr and 1.69 yr^2 basis offsets present in 7 and 9 parameter solutions:
            diffs -= compute_basis_offsets(accRA, accDec, jerkRA, jerkDec)
            # diffs[-4:] are not differences but actually the accRA, accDec, jerkRA, jerkDec parameters, so we need to subtract those off
            diffs[-4:] -= np.array([accRA, accDec, jerkRA, jerkDec])

        return tuple((diffs, errors, chisq, chi2_partials, soltype))
    else:
        return [None] * 9, [None] * 9, None, [None] * 4, soltype


def refit_hip2_object(iad_dir, hip_id, catalog: Table, seven_p_annex: Table = None, nine_p_annex: Table = None, use_parallax=False):
    data = HipparcosRereductionDVDBook()
    data.parse(star_id=hip_id, intermediate_data_directory=iad_dir)

    cat_params, cat_errors, soltype = get_cat_values_hip2(hip_id, catalog)
    cntr_RA, cntr_Dec = cat_params[1:3]
    soltype = soltype.strip()
    fit_degree = {'5': 1, '7': 2, '9': 3}.get(soltype[-1], None)
    # do the fit for seven/nine parameter fits if we have the 7th and 9th parameters.
    accRA_err, accDec_err, jerkRA_err, jerkDec_err = 0, 0, 0, 0
    if seven_p_annex is not None and fit_degree >= 2:
        idx = np.searchsorted(seven_p_annex['hip_id'].data, int(hip_id))  # int(hip_id) strips leading zeroes.
        accRA_err, accDec_err = seven_p_annex[idx][['acc_ra_err', 'acc_dec_err']]
    if nine_p_annex is not None and fit_degree == 3:
        idx = np.searchsorted(nine_p_annex['hip_id'].data, int(hip_id))  # int(hip_id) strips leading zeroes.
        jerkRA_err, jerkDec_err = nine_p_annex[idx][['jerk_ra_err', 'jerk_dec_err']]
    # do the fit
    cat_errors = np.hstack([cat_errors, [accRA_err, accDec_err, jerkRA_err, jerkDec_err]])
    if fit_degree is not None:
        diffs, errors, chisq, chi2_partials = refit_hip_fromdata(data, fit_degree, cntr_RA=cntr_RA, cntr_Dec=cntr_Dec,
                                                                 use_parallax=use_parallax)
        return tuple((diffs, errors - cat_errors, chisq, chi2_partials, soltype))
    else:
        return [None] * 9, [None] * 9, None, [None] * 4, soltype


def refit_hip21_object(iad_dir, hip_id, use_parallax=False):
    data = HipparcosRereductionJavaTool()
    data.parse(star_id=hip_id, intermediate_data_directory=iad_dir)
    fname = glob(os.path.join(iad_dir, '**/', "H" + hip_id.zfill(6) + ".csv"))[0]

    plx, cntr_RA, cntr_Dec, pmRA, pmDec, soltype = get_cat_values_hip21(fname)
    soltype = soltype.strip()
    fit_degree = {'5': 1, '7': 2, '9': 3}.get(soltype[-1], None)
    # For now, just do the 5 parameter sources of Hip2.
    if fit_degree == 1:
        diffs, errors, chisq, chi2_partials = refit_hip_fromdata(data, fit_degree, cntr_RA=cntr_RA, cntr_Dec=cntr_Dec,
                                                                 use_parallax=use_parallax)
        return tuple((diffs, errors, chisq, chi2_partials, soltype))
    else:
        return [None] * 9, [None] * 9, None, [None] * 4, soltype


def get_cat_values_hip1(fname):
    with open(fname) as f:
        lines = f.readlines()
        try:
            pmRA = float(lines[5].split(':')[1].split('P')[0])
            pmDec = float(lines[6].split(':')[1].split('P')[0])
            cntr_RA = Angle(float(lines[2].split(':')[1].split('Right')[0]), unit='degree')
            cntr_Dec = Angle(float(lines[3].split(':')[1].split('Declination')[0]), unit='degree')
            plx = float(lines[4].split(':')[1].split('Trigonometric')[0])
            sol_type = str(lines[7].split(':')[1].split('Code')[0])
        except:
            raise UnboundLocalError('could not read pmRA or pmDec from intermediate data of {0}'.format(fname))
    return plx, cntr_RA, cntr_Dec, pmRA, pmDec, sol_type


def get_cat_values_hip21(fname):
    with open(fname) as f:
        lines = f.readlines()
        try:
            pmRA = float(lines[2].split()[3])
            pmDec = float(lines[2].split()[4])
            cntr_RA = Angle(float(lines[2].split()[0]), unit='degree')
            cntr_Dec = Angle(float(lines[2].split()[1]), unit='degree')
            plx = float(lines[2].split()[2])
            sol_type = str(lines[0].split()[4])
        except:
            raise UnboundLocalError('could not read pmRA or pmDec from intermediate data of {0}'.format(fname))
    return plx, cntr_RA, cntr_Dec, pmRA, pmDec, str(int(sol_type))


def get_cat_values_hip2(hip_id, catalog: Table):
    idx = np.searchsorted(catalog['hip_id'].data, int(hip_id))  # int(hip_id) strips leading zeroes.
    plx, cntr_RA, cntr_Dec, pmRA, pmDec, soltype = catalog[idx]['plx', 'ra', 'dec', 'pmRA', 'pmDec', 'soltype']
    plx_e, cntr_RA_e, cntr_Dec_e, pmRA_e, pmDec_e = catalog[idx]['plx_err', 'ra_err', 'dec_err', 'pmRA_err', 'pmDec_err']
    cntr_RA, cntr_Dec = Angle(cntr_RA, unit=catalog['ra'].unit), Angle(cntr_Dec, unit=catalog['dec'].unit)
    #
    params = [plx, cntr_RA, cntr_Dec, pmRA, pmDec]
    errors = [plx_e, cntr_RA_e, cntr_Dec_e, pmRA_e, pmDec_e]
    return params, errors, str(int(soltype))


def load_hip2_catalog(catalog_path):
    catalog = Table.read(catalog_path, format='ascii')
    catalog = catalog['col1', 'col7', 'col5', 'col6', 'col8', 'col9', 'col12', 'col10', 'col11', 'col13', 'col14', 'col2']
    catalog['col5'] = Angle(Angle(catalog['col5'], unit='rad'), unit='degree')
    catalog['col6'] = Angle(Angle(catalog['col6'], unit='rad'), unit='degree')
    # names from Table G.3 of Hipparcos: The new reduction of the raw data (2007)
    for name, rename in zip(['col1', 'col7', 'col5', 'col6', 'col8', 'col9', 'col12', 'col10', 'col11', 'col13', 'col14', 'col2'],
                            ['hip_id', 'plx', 'ra', 'dec', 'pmRA', 'pmDec', 'plx_err', 'ra_err', 'dec_err', 'pmRA_err', 'pmDec_err', 'soltype']):
        catalog.rename_column(name, rename)
    # sort for quick retrieval of data for any source.
    catalog.sort('hip_id')
    return catalog


def load_hip2_seven_p_annex(path):
    catalog = Table.read(path, format='ascii')
    catalog = catalog['col1', 'col3', 'col4', 'col5', 'col6']
    # names from Table G.5 of Hipparcos: The new reduction of the raw data (2007)
    for name, rename in zip(['col1', 'col3', 'col4', 'col5', 'col6'],
                            ['hip_id', 'acc_ra', 'acc_dec', 'acc_ra_err', 'acc_dec_err']):
        catalog.rename_column(name, rename)
    # sort for quick retrieval of data for any source.
    catalog.sort('hip_id')
    return catalog


def load_hip2_nine_p_annex(path):
    catalog = Table.read(path, format='ascii')
    catalog = catalog['col1', 'col3', 'col4', 'col5', 'col6', 'col7', 'col8', 'col9', 'col10']
    # names from Table G.6 of Hipparcos: The new reduction of the raw data (2007)
    for name, rename in zip(['col1', 'col3', 'col4', 'col5', 'col6', 'col7', 'col8', 'col9', 'col10'],
                            ['hip_id', 'acc_ra', 'acc_dec', 'jerk_ra', 'jerk_dec', 'acc_ra_err', 'acc_dec_err', 'jerk_ra_err', 'jerk_dec_err']):
        catalog.rename_column(name, rename)
    # sort for quick retrieval of data for any source.
    catalog.sort('hip_id')
    return catalog


def load_hip1_dm_annex(catalog_path):
    # load the hip1 doubles and multiples annex for the accelerating systems
    if not os.path.exists(catalog_path):
        warnings.warn('Doubles and multiples annex {0} file not found. Will only fit 5 parameter solutions.'
                      'For 7, 9 parameter validation, please download hip_dm_g.dat from '
                      'https://cdsarc.unistra.fr/ftp/I/239/hip_dm_g.dat.gz (it is not on Vizier as of Sep 16 2020). '
                      'Unzip it, and place it inside the hip1'
                      'intermediate data directory.'.format(catalog_path), UserWarning)
        return None
    catalog = Table.read(catalog_path, format='ascii')
    # column designations are from the Hipparcos and Tycho catalogs, Vol1, Table 2.3.3.
    catalog = catalog['col1', 'col2', 'col3', 'col4', 'col5', 'col7', 'col8', 'col9', 'col10']
    for name, rename in zip(['col1', 'col2', 'col3', 'col4', 'col5', 'col7', 'col8', 'col9', 'col10'],
                            ['hip_id', 'accRA', 'accDec', 'accRA_err', 'accDec_err', 'jerkRA', 'jerkDec', 'jerkRA_err', 'jerkDec_err']):
        catalog.rename_column(name, rename)
    # sort for quick retrieval of data for any source.
    catalog.sort('hip_id')
    return catalog.filled(0)  # this sets missing acc/jerks to 0 and returns a Table instead of a Masked Table


def load_hip1_catalog(catalog_path):
    catalog = Table.read(catalog_path, format='ascii')
    # catalog should be fetched from https://cdsarc.unistra.fr/ftp/I/239/hip_main.dat.gz
    # column designations are from https://cdsarc.unistra.fr/ftp/I/239/ReadMe
    catalog = catalog['col2', 'col12', 'col9', 'col10', 'col13', 'col14', 'col17', 'col15', 'col16', 'col18', 'col19', 'col60']
    for name, rename in zip(['col2', 'col12', 'col9', 'col10', 'col13', 'col14', 'col17', 'col15', 'col16', 'col18', 'col19', 'col60'],
                            ['hip_id', 'plx', 'ra', 'dec', 'pmRA', 'pmDec', 'plx_err', 'ra_err', 'dec_err', 'pmRA_err', 'pmDec_err', 'MultFlag']):
        catalog.rename_column(name, rename)
    # sort for quick retrieval of data for any source.
    catalog.sort('hip_id')
    return catalog.filled(0)  # this sets missing entries to 0 and returns a Table instead of a Masked Table


def compute_basis_offsets(accRA, accDec, jerkRA, jerkDec):
    # account for the 0.81 yr and 1.69 yr^2 basis offsets:
    # See THE HIPPARCOS CATALOGUE DOUBLE AND MULTIPLE SYSTEMS ANNEX by Lennart Lindegren (1997)
    a, b = 0.81, 1.69  # yr^2
    offsets = np.array([0, -a / 2 * accRA, -a / 2 * accDec, -b / 6 * jerkRA, -b / 6 * jerkDec, 0, 0, 0, 0])
    # compute differences between reference parameters and our best fit parameters
    return offsets



"""
Script for looking at d_chi^2/d(a_i) where a_i are the standard astrometric parameters.
"""
import numpy as np
from astropy.coordinates import Angle
from argparse import ArgumentParser
from multiprocessing import Pool
import os

from htof.main import Astrometry
from astropy.table import Table


def chisq_partials(hip_id: str, iad_dir: str, reduction: str, fit_degree=1):
    try:
        astro = Astrometry(reduction, hip_id, iad_dir, central_epoch_ra=1991.25, normed=False,
                           central_epoch_dec=1991.25, format='jyear', fit_degree=fit_degree, use_parallax=False)
        ra_resid = Angle(astro.data.residuals.values * np.sin(astro.data.scan_angle.values), unit='mas')
        dec_resid = Angle(astro.data.residuals.values * np.cos(astro.data.scan_angle.values), unit='mas')
    except FileNotFoundError:
        return np.ones(5, dtype=float) * 10000
    return astro.fitter._chi2_vector(ra_resid.mas, dec_resid.mas)


class ChisqNullTestEngine(object):
    def __init__(self, dirname, reduction):
        self.dirname = dirname
        self.reduction = reduction

    def __call__(self, hip_id):
        result = chisq_partials(hip_id, self.dirname, self.reduction)
        return {'hip_id': hip_id, 'dxdra0': result[0], 'dxddec0': result[1], 'dxdmura': result[2], 'dxdmudec': result[3]}


if __name__ == "__main__":
    parser = ArgumentParser(description='Script for checking the partial derivatives of chisquared. ignores parallax.')
    parser.add_argument("-dir", "--iad-directory", required=True, default=None,
                        help="full path to the intermediate data directory")
    parser.add_argument("-hr", "--hip-reduction", required=True, default=None, type=int,
                        help="integer. 1 for 1997 reduction, 2 for 2007 CD reduction, 21 for 2007 IADT tool.")
    parser.add_argument("-o", "--output-file", required=False, default=None,
                        help="The output filename, with .csv extension. E.g. hip1_refit.csv."
                             "Will default to hip_processid.csv.")
    parser.add_argument("-c", "--cores", required=False, default=1, type=int,
                        help="Number of cores to use. Default is 1.")
    parser.add_argument("-in", "--in-list", required=True, type=str,
                        help="In list of hip_ids to check. Must be a standard text file OR a .csv file "
                             "with a header called hip_id which gives the hip_id of the sources you want to fit")
    parser.add_argument("--debug", action='store_true', default=False, required=False,
                        help='If true, this will run the refit test on only 500 sources. Useful to check for '
                             'filepath problems before running the full test on all ~100000 sources.')

    args = parser.parse_args()
    if '.csv' not in args.in_list:
        hip_ids = np.genfromtxt(args.in_list).flatten().astype(str)
    else:
        hip_ids = np.array(Table.read(args.in_list)['hip_id'].data, dtype=str).flatten()

    if args.output_file is None:
        output_file = 'hip' + args.hip_reduction + '_chisq_null_test' + (str)(os.getpid()) + '.csv'
    else:
        output_file = args.output_file

    engine = ChisqNullTestEngine(args.iad_directory, 'hip' + str(args.hip_reduction))
    if args.debug:
        hip_ids = hip_ids[:500]
    print('will check {0} total hip {1} objects'.format(len(hip_ids), str(args.hip_reduction)))
    print('will save output table at', output_file)
    try:
        pool = Pool(args.cores)
        data_outputs = pool.map(engine, hip_ids)
        out = Table(data_outputs)
        out.sort('hip_id')
        out.write(output_file, overwrite=True)
    finally: # To make sure processes are closed in the end, even if errors happen
        pool.close()
        pool.join()



import pandas as pd
import numpy as np
import pytest
import mock
import os
import tempfile
from ast import literal_eval

from astropy.table import Table
from htof.parse import HipparcosOriginalData, HipparcosRereductionDVDBook,\
    GaiaData, DataParser, GaiaDR2, GaiaeDR3, DecimalYearData, HipparcosRereductionJavaTool, digits_only, \
    match_filename, get_nparam
from htof.parse import calc_inverse_covariance_matrices


class TestHipparcosOriginalData:
    def test_parse(self, hip_id='027321'):
        test_data_directory = os.path.join(os.getcwd(), 'htof/test/data_for_tests/Hip1')
        data = HipparcosOriginalData()
        data.parse(star_id=hip_id,
                   intermediate_data_directory=test_data_directory,
                   data_choice='FAST')
        assert len(data) == 32
        assert np.isclose(data._epoch[0], 1990.005772)
        assert np.isclose(np.sin(data.scan_angle[0]), -0.9053, rtol=.01)
        assert np.isclose(data._epoch[17], 1990.779865)
        assert np.isclose(np.sin(data.scan_angle[17]), 0.3633, rtol=.01)
        assert np.isclose(data.along_scan_errs.values[0], 2.21)
        data.parse(star_id=hip_id,
                   intermediate_data_directory=test_data_directory,
                   data_choice='NDAC')
        assert len(data) == 34
        assert np.isclose(data._epoch[1], 1990.005386)
        assert np.isclose(np.sin(data.scan_angle[1]), -0.9051, rtol=.01)
        assert np.isclose(data._epoch[10], 1990.455515)
        assert np.isclose(np.sin(data.scan_angle[10]), 0.7362, rtol=.01)
        data.parse(star_id=hip_id,
                   intermediate_data_directory=test_data_directory,
                   data_choice='MERGED')
        assert len(data) == 34
        assert np.isclose(data._epoch[0], 1990.005386)
        assert np.isclose(np.sin(data.scan_angle[0]), -0.9053, atol=.001)
        assert np.isclose(data._epoch[5], 1990.455515)
        assert np.isclose(np.sin(data.scan_angle[5]), 0.7364, atol=.001)
        assert np.isclose(data.along_scan_errs[5], 2.0814, atol=.0001)
        assert np.isclose(data.residuals[5], 1.1021, atol=.0001)
        data.parse(star_id=hip_id,
                   intermediate_data_directory=test_data_directory,
                   data_choice='BOTH')
        assert len(data) == 66

    @pytest.mark.integration
    def test_concatenation(self, hip_id='027321'):
        test_data_directory = os.path.join(os.getcwd(), 'htof/test/data_for_tests/Hip1')
        data = HipparcosOriginalData()
        data.parse(star_id=hip_id,
                   intermediate_data_directory=test_data_directory,
                   data_choice='FAST')
        data.calculate_inverse_covariance_matrices()
        covars = data.inverse_covariance_matrix
        data += data
        data.calculate_inverse_covariance_matrices()
        assert np.allclose(covars, data.inverse_covariance_matrix[:len(covars)])
        assert np.allclose(covars, data.inverse_covariance_matrix[len(covars):])

    def test_parse_IA3_eq_zero(self, hip_id='004391'):
        test_data_directory = os.path.join(os.getcwd(), 'htof/test/data_for_tests/Hip1')
        data = HipparcosOriginalData()
        data.parse(star_id=hip_id,
                   intermediate_data_directory=test_data_directory,
                   data_choice='NDAC')
        # IA4 is larger than IA3 -> more precise answer from IA7/IA4 (differs by .0103538462)
        assert np.isclose(data._epoch[38], 1992.9142, rtol=1e-8)
        # IA3 is larger than IA4 -> more precise answer from IA6/IA3 (differs by .0001434115)
        assert np.isclose(data._epoch[5], 1990.4227657727, rtol=1e-8)
        # IA3 is exactly 0 which would result in NaN. Must use IA7/IA4 to get a valid result. 
        # IA4 is negative. This case thus also checks that the absolute value is correctly used when comparing IA3 and IA4.
        assert np.isclose(data._epoch[30], 1992.407684232, rtol=1e-8)

    def test_merged_parse_removes_flagged_observations(self):
        test_data_directory = os.path.join(os.getcwd(), 'htof/test/data_for_tests/Hip1')
        data = HipparcosOriginalData()
        data.parse(star_id='999999',
                   intermediate_data_directory=test_data_directory,
                   data_choice='MERGED')
        assert len(data) == 32

    def test_raises_exception_on_bad_data_choice(self):
        test_data_directory = os.path.join(os.getcwd(), 'htof/test/data_for_tests/Hip1')
        data = HipparcosOriginalData()
        with pytest.raises(Exception):
            data.parse(star_id='027321',
                       intermediate_data_directory=test_data_directory,
                       data_choice='something')


class TestHipparcosRereductionDVDBook:
    def test_error_inflation_factor(self):
        u = 0.875291 # D. Michalik et al. 2014 Q factor for Hip 27321, calculated by hand
        assert np.isclose(HipparcosRereductionDVDBook.error_inflation_factor(111, 5, -1.81), u, rtol=1e-5)

    def test_parse(self):
        test_data_directory = os.path.join(os.getcwd(), 'htof/test/data_for_tests/Hip2')
        data = HipparcosRereductionDVDBook()
        data.parse(star_id='027321', intermediate_data_directory=test_data_directory)
        nu = 111 - 5
        Q = nu * (np.sqrt(2/(9*nu))*-1.81 + 1 - 2/(9*nu))**3  # D. Michalik et al. 2014 Q factor for Hip 27321
        # Note that F2 = -1.81 in the CD intermediate data, while F2 = -1.63 on Vizier.
        u = np.sqrt(Q/nu)  # error inflation factor.
        assert len(data) == 111
        assert np.isclose(data._epoch[0], 1990.005)
        assert np.isclose(np.sin(data.scan_angle[0]), -0.9065, rtol=.01)
        assert np.isclose(data.along_scan_errs.values[0], 0.81 * u)
        assert np.isclose(data._epoch[84], 1991.952)
        assert np.isclose(np.sin(data.scan_angle[84]), -0.8083, rtol=.01)

    def test_reject_obs(self):
        test_data_directory = os.path.join(os.getcwd(), 'htof/test/data_for_tests/Hip2')
        data = HipparcosRereductionDVDBook()
        data.parse(star_id='27321', intermediate_data_directory=test_data_directory)
        assert len(data) == 111 - 0
        assert data.rejected_epochs == []

        data.parse(star_id='84', intermediate_data_directory=test_data_directory)
        assert len(data) == 96 - 1
        assert np.allclose(np.sort(data.rejected_epochs), [69])

        data.parse(star_id='70', intermediate_data_directory=test_data_directory)
        assert len(data) == 112 - 5
        assert np.allclose(np.sort(data.rejected_epochs), [36, 55, 59, 64, 105])

        data.parse(star_id='40', intermediate_data_directory=test_data_directory)
        assert len(data) == 134 - 2
        assert np.allclose(np.sort(data.rejected_epochs), [6, 53])

        data.parse(star_id='072477', intermediate_data_directory=test_data_directory)
        assert len(data) == 64 - 0
        assert data.rejected_epochs == []



class TestHipparcosRereductionJavaTool:
    test_data_directory = os.path.join(os.getcwd(), 'htof/test/data_for_tests/Hip21')

    def test_parse(self):
        data = HipparcosRereductionJavaTool()
        data.parse(star_id='27321', intermediate_data_directory=self.test_data_directory)
        u = 1  # Error inflation factor
        assert len(data) == 111
        assert np.isclose(data._epoch[0], 1990.0055)
        assert np.isclose(np.sin(data.scan_angle[0]), -0.9050, rtol=.01)
        assert np.isclose(data.along_scan_errs.values[0], 0.80 * u)
        assert np.isclose(data._epoch[84], 1991.9523)
        assert np.isclose(np.sin(data.scan_angle[84]), -0.8083, rtol=.01)

    def test_outlier_reject(self):
        data = HipparcosRereductionJavaTool()
        data.parse(star_id='27100', intermediate_data_directory=self.test_data_directory)
        assert len(data) == 147 - 2  # num entries - num outliers
        # outliers are marked with negative AL errors. Assert outliers are gone.
        assert np.all(data.along_scan_errs > 0)


class TestDataParser:
    def test_parse_raises_file_not_found_error(self):
        with pytest.raises(FileNotFoundError):
            test_data_directory = os.path.join(os.getcwd(), 'htof/test/data_for_tests/Hip2')
            data = HipparcosRereductionDVDBook()
            data.parse(star_id='12gjas2',
                       intermediate_data_directory=test_data_directory)

    @mock.patch('htof.parse.glob.glob', return_value=['path/027321.dat', 'path/027321.dat'])
    def test_parse_raises_error_on_many_files_found(self, fake_glob):
        test_data_directory = os.path.join(os.getcwd(), 'htof/test/data_for_tests/Hip2')
        data = HipparcosRereductionDVDBook()
        with pytest.raises(FileNotFoundError):
            data.parse(star_id='027321',
                       intermediate_data_directory=test_data_directory)

    @mock.patch('htof.parse.pd.read_csv', return_value=None)
    @mock.patch('htof.parse.glob.glob', return_value=['path/127321.dat', 'path/27321.dat'])
    def test_read_on_couple_similar_filepaths(self, fake_glob, fake_load):
        test_data_directory = os.path.join(os.getcwd(), 'path/')
        DataParser().read_intermediate_data_file('27321', test_data_directory, None, None, None)
        fake_load.assert_called_with('path/27321.dat', sep=None, skiprows=None, header=None, engine='python')

    @mock.patch('htof.parse.pd.read_csv', return_value=None)
    @mock.patch('htof.parse.glob.glob')
    def test_read_on_many_filepaths(self, fake_glob, fake_load):
        test_data_directory = os.path.join(os.getcwd(), 'path/')
        fake_glob.return_value = ['/fake/path/1232.dat', '/fake/path/23211.dat', '/fake/path/232.dat']
        DataParser().read_intermediate_data_file('232', test_data_directory, None, None, None)
        fake_load.assert_called_with('/fake/path/232.dat', sep=None, skiprows=None, header=None, engine='python')
        # Test hip2 style names
        fake_glob.return_value = ['H075/HIP075290.d', 'H075/HIP075293.d', 'H107/HIP107529.d', 'H007/HIP007529.d', 'H117/HIP117529.d']
        DataParser().read_intermediate_data_file('7529', test_data_directory, None, None, None)
        fake_load.assert_called_with('H007/HIP007529.d', sep=None, skiprows=None, header=None, engine='python')

    def test_match_filename(self):
        paths = ['/hip/24374.txt', 'hip/43749.txt', 'hip/43740.txt', 'hip/4374.txt']
        assert 'hip/4374.txt' == match_filename(paths, '4374')[0]

    @mock.patch('htof.parse.pd.read_csv', return_value=None)
    @mock.patch('htof.parse.glob.glob')
    def test_read_on_irregular_paths(self, fake_glob, fake_load):
        test_data_directory = os.path.join(os.getcwd(), 'path/')
        fake_glob.return_value = ['/fake/path/SSADSx1232.dat', '/fake/path/WDASxs23211.dat', '/fake/path/OMGH232.dat']
        DataParser().read_intermediate_data_file('232', test_data_directory, None, None, None)
        fake_load.assert_called_with('/fake/path/OMGH232.dat', sep=None, skiprows=None, header=None, engine='python')

    def test_len(self):
        assert len(DataParser()) == 0
        assert len(DataParser(epoch=pd.DataFrame(np.arange(4)))) == 4


def test_convert_dates_to_jd():
    epoch = pd.DataFrame(data=[1990.0, 1990.25], index=[5, 6])
    parser = DecimalYearData(epoch=epoch)
    jd_epochs = parser.julian_day_epoch()
    assert np.isclose(jd_epochs[0], 2447892.5)
    assert np.isclose(jd_epochs[1], 2447892.5 + 0.25*365.25)


def test_call_jd_dates_hip():
    parser = DecimalYearData()
    parser._epoch = pd.DataFrame(data=[1990.0, 1990.25], index=[5, 6])
    jd_epochs = parser.julian_day_epoch()
    assert np.isclose(jd_epochs[0], 2447892.5)
    assert np.isclose(jd_epochs[1], 2447892.5 + 0.25*365.25)


def test_call_jd_dates_gaia():
    parser = DataParser()
    parser._epoch = pd.DataFrame(data=[2447892.5, 2447893], index=[5, 6])
    jd_epochs = parser.julian_day_epoch()
    assert np.isclose(jd_epochs[0], 2447892.5)
    assert np.isclose(jd_epochs[1], 2447893)


def test_trim_gaia_data():
    parser = GaiaData()
    datemin, datemax = 3, 5
    epochs = pd.DataFrame(data=[datemin - 1, datemin, datemax, datemax + 1], index=[3, 4, 5, 6])
    data = pd.DataFrame(data=[datemin - 1, datemin, datemax, datemax + 1], index=[3, 4, 5, 6])
    data = parser.trim_data(data, epochs, datemin, datemax)
    assert np.allclose(data.values.flatten(), [datemin, datemax])


@mock.patch('htof.parse.calc_inverse_covariance_matrices', return_value=np.array([np.ones((2, 2))]))
def test_calculate_inverse_covariances(mock_cov_matrix):
    parser = DataParser()
    parser.calculate_inverse_covariance_matrices()
    assert np.allclose(parser.inverse_covariance_matrix[0], np.ones((2, 2)))


class TestParseGaiaData:
    @pytest.mark.integration
    def test_parse_all_epochs(self):
        test_data_directory = os.path.join(os.getcwd(), 'htof/test/data_for_tests/GaiaDR2/IntermediateData')
        data = GaiaData(max_epoch=np.inf, min_epoch=-np.inf)
        data.parse(intermediate_data_directory=test_data_directory,
                   star_id='049699')
        assert len(data._epoch) == 72
        assert np.isclose(data._epoch[0], 2456951.7659301492)
        assert np.isclose(data.scan_angle[0], -1.8904696884345342)
        assert np.isclose(data._epoch[70], 2458426.7784441216)
        assert np.isclose(data.scan_angle[70], 2.821818345385301)

    @pytest.mark.integration
    def test_parse_selects_valid_epochs(self):
        test_data_directory = os.path.join(os.getcwd(), 'htof/test/data_for_tests/GaiaDR2/IntermediateData')
        data = GaiaDR2(max_epoch=2458426.7784441218, min_epoch=2457143.4935643710)
        data.DEAD_TIME_TABLE_NAME = None  # Do not reject dead times for this test
        data.parse(intermediate_data_directory=test_data_directory,
                   star_id='049699')

        assert len(data._epoch) == 68
        assert np.isclose(data._epoch.iloc[0], 2457143.4935643715)
        assert np.isclose(data.scan_angle.iloc[0], -0.3066803677989655)
        assert np.isclose(data._epoch.iloc[67], 2458426.7784441216)
        assert np.isclose(data.scan_angle.iloc[67], 2.821818345385301)

    @pytest.mark.integration
    def test_DR2parse_removes_dead_times(self):
        # tests the dead time table against a fake source which has the first observation fall in a data gap
        # and has the second observation not fall into any data gap.
        test_data_directory = os.path.join(os.getcwd(), 'htof/test/data_for_tests/GaiaDR2/IntermediateData')
        data = GaiaDR2(max_epoch=np.inf, min_epoch=-np.inf)
        data.parse(intermediate_data_directory=test_data_directory,
                   star_id='gaiatestsource01')

        assert len(data._epoch) == 1
        assert np.isclose(data._epoch.iloc[0], 2456893.28785)
        assert np.isclose(data.scan_angle.iloc[0], -1.7804696884345342)

    @pytest.mark.integration
    def test_eDR3parse_removes_dead_times(self):
        # tests the dead time table against a fake source which has the first observation fall in a data gap
        # and has the second observation not fall into any data gap.
        test_data_directory = os.path.join(os.getcwd(), 'htof/test/data_for_tests/GaiaeDR3/IntermediateData')
        data = GaiaeDR3(max_epoch=np.inf, min_epoch=-np.inf)
        data.parse(intermediate_data_directory=test_data_directory,
                   star_id='gaiatestsource01')

        assert len(data._epoch) == 1
        assert np.isclose(data._epoch.iloc[0], 2456893.28785)
        assert np.isclose(data.scan_angle.iloc[0], -1.7804696884345342)


def test_write_with_missing_info():
    data = DataParser(scan_angle=np.arange(3), epoch=np.arange(1991, 1994),
                      residuals=np.arange(2, 5),
                      inverse_covariance_matrix=None,
                      along_scan_errs=None)
    with tempfile.TemporaryDirectory() as tmp_dir:
        data.write(os.path.join(tmp_dir, 'out.csv'))
        t = Table.read(os.path.join(tmp_dir, 'out.csv'))
        assert np.allclose(t['residuals'], data.residuals)
        assert np.allclose(t['julian_day_epoch'], data.julian_day_epoch())
        assert np.allclose(t['scan_angle'], data.scan_angle)
        assert len(t.colnames) == 5


def test_write():
    data = DataParser(scan_angle=np.arange(3), epoch=np.arange(1991, 1994),
                      residuals=np.arange(2, 5),
                      inverse_covariance_matrix=np.array([[1, 2], [3, 4]]) * np.ones((3, 2, 2)),
                      along_scan_errs=np.arange(3, 6))
    with tempfile.TemporaryDirectory() as tmp_dir:
        data.write(os.path.join(tmp_dir, 'out.csv'))
        t = Table.read(os.path.join(tmp_dir, 'out.csv'))
        assert np.allclose(t['residuals'], data.residuals)
        assert np.allclose(t['julian_day_epoch'], data.julian_day_epoch())
        assert np.allclose(t['scan_angle'], data.scan_angle)
        assert np.allclose(t['along_scan_errs'], data.along_scan_errs)
        icovs = [np.array(literal_eval(icov)) for icov in t['icov']]
        assert np.allclose(icovs, data.inverse_covariance_matrix)
        assert len(t.colnames) == 5


def test_calculating_covariance_matrices():
    scan_angles = pd.DataFrame(data=np.linspace(0, 2 * np.pi, 5))
    icovariances = calc_inverse_covariance_matrices(scan_angles, cross_scan_along_scan_var_ratio=10)
    assert len(icovariances) == len(scan_angles)
    # calculate the covariance matrices so that we can compute the scan angle from the matrix.
    covariances = np.zeros_like(icovariances)
    for i in range(len(icovariances)):
        covariances[i] = np.linalg.pinv(icovariances[i])
    assert np.allclose(covariances[-1], covariances[0])  # check 2pi is equivalent to 0.
    assert np.allclose(covariances[0], np.array([[10, 0], [0, 1]]))  # angle of 0 has AL parallel with DEC.
    for cov_matrix, scan_angle in zip(covariances, scan_angles.values.flatten()):
        assert np.isclose(scan_angle % np.pi, angle_of_short_axis_of_error_ellipse(cov_matrix) % np.pi)
        # modulo pi since the scan angle and angle of short axis could differ in sign from one another.


def test_concatenating_data():
    data = DataParser(scan_angle=np.arange(3), epoch=np.arange(1991, 1994),
                      residuals=np.arange(2, 5),
                      inverse_covariance_matrix=np.array([[1, 2], [3, 4]]) * np.ones((3, 2, 2)),
                      along_scan_errs=np.arange(3, 6))
    new_data = sum([data, data])
    assert np.allclose(new_data.scan_angle, [*data.scan_angle, *data.scan_angle])
    assert np.allclose(new_data.residuals, [*data.residuals, *data.residuals])
    assert len(new_data.inverse_covariance_matrix) == 2*len(data.inverse_covariance_matrix)
    data += data
    assert np.allclose(new_data.scan_angle, data.scan_angle)
    assert np.allclose(new_data.residuals, data.residuals)
    data.calculate_inverse_covariance_matrices()


def test_add_to_empty():
    data = DataParser(scan_angle=np.arange(3), epoch=pd.DataFrame(np.arange(1991, 1994)),
                      residuals=np.arange(2, 5),
                      inverse_covariance_matrix=np.array([[1, 2], [3, 4]]) * np.ones((3, 2, 2)),
                      along_scan_errs=np.arange(3, 6))
    new_data = DataParser()
    new_data += data
    assert np.allclose(new_data.scan_angle, data.scan_angle)
    assert len(new_data) == len(data)


def test_concatenating_data_with_missing():
    data = DataParser(scan_angle=np.arange(3), epoch=pd.DataFrame(np.arange(1991, 1994)),
                      residuals=np.arange(2, 5))
    new_data = sum([data, data])
    assert np.allclose(new_data.scan_angle, [*data.scan_angle, *data.scan_angle])
    assert np.allclose(new_data.residuals, [*data.residuals, *data.residuals])
    data += data
    assert np.allclose(new_data.scan_angle, data.scan_angle)
    assert np.allclose(new_data.residuals, data.residuals)


@pytest.mark.integration
def test_two_concatenate_decyear_and_jd():
    data = DecimalYearData(scan_angle=np.arange(3), epoch=pd.DataFrame(np.arange(1991, 1994)),
                           residuals=np.arange(2, 5))
    data2 = DataParser(scan_angle=np.arange(3), epoch=pd.DataFrame(np.arange(2456951, 2456954)),
                       residuals=np.arange(2, 5))
    data3 = data + data2
    assert np.allclose(data3.julian_day_epoch()[:len(data)], data.julian_day_epoch())
    assert np.allclose(data3.julian_day_epoch()[len(data):], data2.julian_day_epoch())


def angle_of_short_axis_of_error_ellipse(cov_matrix):
    vals, vecs = np.linalg.eigh(cov_matrix)
    # Compute "tilt" of ellipse using first eigenvector
    x, y = vecs[:, 0]
    theta = np.arctan2(y, x) - np.pi/2
    return theta


def test_digits_only():
    assert '987978098098098' == digits_only("sdkjh987978asd098as0980a98sd")


def test_get_nparam():
    assert 5 == get_nparam('23115')
    assert 5 == get_nparam('95')
    assert 7 == get_nparam('97')



import numpy as np
from numpy.polynomial.polynomial import Polynomial
from scipy.special import factorial


def taylorvander(x, deg):
    """
    Taylor series analog for the Vandermonde matrix of given degree.
    Returns the Vandermonde matrix of degree `deg` and sample points
    `x`. The Vandermonde matrix is defined by
    .. math:: V[..., i] = x^i / i!,
    where `0 <= i <= deg`. The leading indices of `V` index the elements of
    `x` and the last index is the power of `x`.
    Parameters
    ----------
    x : array_like
        Array of points. The dtype is converted to float64 or complex128
        depending on whether any of the elements are complex. If `x` is
        scalar it is converted to a 1-D array.
    deg : int
        Degree of the resulting matrix.
    Returns
    -------
    vander : ndarray.
        The Vandermonde matrix. The shape of the returned matrix is
        ``x.shape + (deg + 1,)``, where the last index is the power of `x`.
        The dtype will be the same as the converted `x`.
        See Also
        --------
        numpy.polynomial.polynomial.polyvander
    Examples
    --------
    >>> x = np.array([1, 2, 3, 5])
    >>> deg = 2
    >>> taylorvander(x, deg)
    array([[ 1.,  1.,  0.5],
           [ 1.,  2.,   2.],
           [ 1.,  3.,  4.5],
           [ 1.,  5., 12.5]])

    """
    return np.polynomial.polynomial.polyvander(x, deg) / factorial(np.arange(deg + 1))


class TaylorSeries(Polynomial):
    def __init__(self, coef, domain=None, window=None):
        mp = factorial(np.arange(len(coef)))
        super(TaylorSeries, self).__init__(coef / mp, domain=domain, window=window)



import numpy as np
from astropy.time import Time
from astropy import units

from htof.sky_path import earth_ephemeris, earth_sun_l2_ephemeris, epoch_topocentric_coordinates, parallactic_motion


def test_earth_ephemeris():
    perihelion, aphelion = Time(['2019-01-02T15:30:00.00', '2019-07-04T01:40:00'], format='isot', scale='utc')
    assert np.allclose(np.sum(earth_ephemeris([perihelion.jyear, aphelion.jyear])**2, axis=0),
                       [0.9832730, 1.0166664], atol=0.004)


def test_earth_sun_l2_ephemeris():
    t = Time('2019-01-02T15:30:00.00', format='isot', scale='utc')
    assert np.allclose(np.sqrt(np.sum(earth_ephemeris(t)**2)) * 1.511 / 1.496,
                       np.sqrt(np.sum(earth_sun_l2_ephemeris(t)**2)))


def test_epoch_topocentric_coordinates():
    alphadeg = np.random.random() * 360 - 180  # random float between -180 and 180
    deltadeg = np.random.random() * 180 - 90  # random float between -90 and 90
    mura = 0  # mas/yr
    parallax = 1  # mas
    mudec = 0  # mas/yr
    vrad = 0  # km/s
    refepoch = Time(1991.25, format='decimalyear', scale='tcb').jyear
    times = np.linspace(refepoch - 1, refepoch + 1, num=100)

    # alpha and delta are the radian values of the stars central coordinate on the sky.
    alpha = (alphadeg * units.degree).to(units.rad).value
    delta = (deltadeg * units.degree).to(units.rad).value

    f_ephem = earth_ephemeris
    ra, dec = epoch_topocentric_coordinates(alpha, delta, parallax, mura, mudec, vrad, times, refepoch, f_ephem)[:2]
    ra2, dec2 = epoch_topocentric_coordinates(alpha, delta, 2 * parallax, mura, mudec, vrad, times, refepoch, f_ephem)[:2]
    assert np.allclose([ra2 - alpha, dec2 - delta], 2 * np.array([ra - alpha, dec - delta]), rtol=1E-5)


def test_parallactic_motion():
    alphadeg = 20
    deltadeg = 20
    parallax = 1000  # mas
    refepoch = Time(1991.25, format='decimalyear', scale='tcb').jyear
    times = np.linspace(refepoch - 1, refepoch + 1, num=100)
    ra, dec = parallactic_motion(times, alphadeg, deltadeg, 'degree', refepoch, earth_ephemeris, parallax)
    ra2, dec2 = parallactic_motion(times, alphadeg, deltadeg, 'degree', refepoch, earth_ephemeris, 2 * parallax)
    assert np.allclose([ra2, dec2], 2 * np.array([ra, dec]), rtol=1E-5)



from htof.parse import HipparcosOriginalData, HipparcosRereductionDVDBook
from htof.fit import AstrometricFitter
from htof.sky_path import parallactic_motion, earth_ephemeris
from astropy import time
from astropy.coordinates import Angle
import numpy as np
from argparse import ArgumentParser
import os


def parse(datacsv, data, data_choice='MERGED', perturb=True):
    if (data_choice is not 'NDAC') and (data_choice is not 'FAST') and (data_choice is not 'MERGED'):
        raise ValueError('data choice has to be either NDAC or FAST or MERGED.')
    datacsv = data._fix_unnamed_column(datacsv)
    if perturb:
        fourdig = ['IA3', 'IA4', 'IA5', 'IA6', 'IA7']
        twodig = ['IA8', 'IA9']
        threedig = ['IA10']
        for cols, digits in zip([fourdig, threedig, twodig], [4, 3, 2]):
            datacsv[cols] = datacsv[cols].to_numpy() + np.random.uniform(-.5 / (10**digits), .5 / (10**digits),
                                                                         datacsv[cols].to_numpy().shape)
    datacsv = data._select_data(datacsv, data_choice)
    # compute scan angles and observations epochs according to van Leeuwen & Evans 1997, eq. 11 & 12.
    data.scan_angle = np.arctan2(datacsv['IA3'], datacsv['IA4'])  # unit radians, arctan2(sin, cos)
    # Use the larger denominator when computing the epoch offset.
    # This increases numerical precision and avoids NaNs if one of the two fields (IA3, IA4) is exactly zero.
    data._epoch = 1991.25 + (datacsv['IA6'] / datacsv['IA3']).where(abs(datacsv['IA3']) > abs(datacsv['IA4']),
                                                              (datacsv['IA7'] / datacsv['IA4']))
    data.residuals = datacsv['IA8']  # unit milli-arcseconds (mas)
    data.along_scan_errs = datacsv['IA9']  # unit milli-arcseconds
    return data


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    parser = ArgumentParser(description='Enter a hip id (string)')
    parser.add_argument("-i", "--hip-id", required=True, default='100564', #1276 is bad
                        help="hip id of the target")
    args = parser.parse_args()

    hip_id = args.hip_id

    data_choice = 'MERGED'
    iad_dir = '/home/gmbrandt/Documents/hipparcosOriginalIntermediateData/'
    # set accRA, accDec to 0 to change to a five parameter model which generates the data.
    accRA = 0#-23.07
    accDec = 0#26.31

    # fit_degree (degree of the model which is fit to the data, i.e. what HTOF uses to fit to the motion)
    # fit_degree = 1 is 5 parameter, fit_degree = 2 is 7 parameter.
    fit_degree = 1
    use_parallax = False
    perturb = True # Whether or not to randomly perturb the digit beyond the last significant digit of the intermediate data.
    ###### end of user params

    with open(os.path.join(iad_dir, hip_id + '.txt')) as f:
        lines = f.readlines()
        try:
            cntr_ra = Angle(float(lines[2].split(':')[1].split('Right')[0]), unit='degree')
            cntr_dec = Angle(float(lines[3].split(':')[1].split('Declination')[0]), unit='degree')
            plx = float(lines[4].split(':')[1].split('Trig')[0])
            pmRA = float(lines[5].split(':')[1].split('Proper')[0])
            pmDec = float(lines[6].split(':')[1].split('Proper')[0])
        except:
            raise UnboundLocalError('could not read pmRA or pmDec from intermediate data of {0}'.format(fname))

    data = HipparcosOriginalData()
    # manually loading file
    datacsv = data.read_intermediate_data_file(hip_id, iad_dir, skiprows=10, header='infer', sep='\s*\|\s*')
    pmRAs, pmDecs = [], []
    for i in range(100):
        # parsing with a pertubation
        data = parse(datacsv.copy(), data, data_choice='MERGED', perturb=perturb)
        #data.parse(star_id=hip_id, intermediate_data_directory=iad_dir,
        #           data_choice=data_choice)
        data.calculate_inverse_covariance_matrices()
        mas_to_degree = 1./60/60/1000

        jyear_epoch = time.Time(data.julian_day_epoch(), format='jd', scale='tcb').jyear
        # note that ra_motion and dec_motion are in degrees here.
        # generate sky path
        year_epochs = jyear_epoch - time.Time(1991.25, format='decimalyear', scale='tcb').jyear
        ra_ref = Angle(pmRA * mas_to_degree * year_epochs, unit='degree')
        dec_ref = Angle(pmDec * mas_to_degree * year_epochs, unit='degree')
        # generate parallax motion
        ra_motion, dec_motion = parallactic_motion(jyear_epoch, cntr_ra.degree, cntr_dec.degree, 'degree',
                                                   time.Time(1991.25, format='decimalyear', scale='tcb').jyear,
                                                   ephemeris=earth_ephemeris)  # Hipparcos was in a geostationary orbit.
        if use_parallax:
            ra_ref += Angle(ra_motion * plx, unit='degree')
            dec_ref += Angle(dec_motion * plx, unit='degree')
        #acceleration terms
        ra_ref += Angle(1/2*accRA * mas_to_degree * (year_epochs**2 - 0.81), unit='degree')
        dec_ref += Angle(1/2*accDec * mas_to_degree * (year_epochs**2 - 0.81), unit='degree')
        #
        ra_model = 1.*ra_ref
        dec_model = 1.*dec_ref

        ra_resid = Angle(data.residuals.values * np.sin(data.scan_angle.values), unit='mas')
        dec_resid = Angle(data.residuals.values * np.cos(data.scan_angle.values), unit='mas')
        ra_ref += ra_resid
        dec_ref += dec_resid
        # instantiate fitter
        fitter = AstrometricFitter(data.inverse_covariance_matrix, year_epochs,
                                   use_parallax=use_parallax, fit_degree=fit_degree,
                                   parallactic_pertubations={'ra_plx': Angle(ra_motion, 'degree').mas,
                                                             'dec_plx': Angle(dec_motion, 'degree').mas})
        # fit
        fit_coeffs = fitter.fit_line(ra_ref.mas, dec_ref.mas)
        pmRAs.append(fit_coeffs[2])
        pmDecs.append(fit_coeffs[3])
        #print(pmRA, pmDec, 'original pmRA and pmDec')
        #print(fit_coeffs[-2:], 'best fit proper motions')

        #if use_parallax:
        #    print(fit_coeffs[0], ' = best fit parallax angle (mas)')
        #    print(plx, ' = input parallax angle (mas)')

    fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(10, 6)) # width by length
    ax1.hist(np.array(pmRAs) - pmRA, label=r'$\Delta \mu_{RA}$')
    ax2.hist(np.array(pmDecs) - pmDec, label=r'$\Delta \mu_{DEC}$')
    ax1.legend(loc='best')
    ax2.legend(loc='best')
    ax1.set_xlabel(r'$\Delta \mu_{RA}$')
    ax2.set_xlabel(r'$\Delta \mu_{DEC}$')
    ax1.set_ylabel('Count')
    ax2.set_ylabel('Count')
    plt.title('Discrepancies w/ respect to catalog PMs for Hip {0}'.format(hip_id))
    #plt.savefig('/home/gmbrandt/Downloads/discrep_{0}.pdf'.format(hip_id))
    plt.show()



"""
Driver script for htof.
The Astrometry class is what a user should use to both parse and fit the intermediate astrometric data.

Author: G. Mirek Brandt
"""

from astropy.time import Time
from astropy.coordinates import Angle
import warnings

from htof.fit import AstrometricFitter
from htof.parse import GaiaeDR3, GaiaDR2, GaiaData
from htof.parse import HipparcosOriginalData, HipparcosRereductionJavaTool, HipparcosRereductionDVDBook
from htof.sky_path import parallactic_motion, earth_ephemeris, earth_sun_l2_ephemeris


class Astrometry(object):
    parsers = {'gaiaedr3': GaiaeDR3, 'gaiadr2': GaiaDR2, 'gaia': GaiaData, 'hip21': HipparcosRereductionJavaTool,
               'hip1': HipparcosOriginalData, 'hip2': HipparcosRereductionDVDBook}
    ephemeri = {'gaiadr2': earth_sun_l2_ephemeris, 'gaia': earth_sun_l2_ephemeris,
                'hip1': earth_ephemeris, 'hip2': earth_ephemeris, 'hip21': earth_ephemeris}

    def __init__(self, data_choice, star_id, intermediate_data_directory, fitter=None, data=None,
                 central_epoch_ra=0, central_epoch_dec=0, format='jd', fit_degree=1,
                 use_parallax=False, central_ra=None, central_dec=None, normed=False):

        if data is None:
            DataParser = self.parsers[data_choice.lower()]
            data = DataParser()
            data.parse(star_id=star_id,
                       intermediate_data_directory=intermediate_data_directory)
            data.calculate_inverse_covariance_matrices()

        parallactic_pertubations = None
        if use_parallax and isinstance(central_ra, Angle) and isinstance(central_dec, Angle):
            if central_epoch_dec != central_epoch_ra:
                warnings.warn('central_epoch_dec != central_epoch_ra. '
                              'Using central_epoch_ra as the central_epoch to compute the parallax motion.',
                              UserWarning)    # pragma: no cover
            ra_motion, dec_motion = parallactic_motion(Time(data.julian_day_epoch(), format='jd').jyear,
                                                       central_ra.mas, central_dec.mas, 'mas',
                                                       Time(central_epoch_ra, format=format).jyear,
                                                       ephemeris=self.ephemeri[data_choice.lower()])
            parallactic_pertubations = {'ra_plx': ra_motion, 'dec_plx': dec_motion}

        if fitter is None and data is not None:
            fitter = AstrometricFitter(inverse_covariance_matrices=data.inverse_covariance_matrix,
                                       epoch_times=Time(Time(data.julian_day_epoch(), format='jd'), format=format).value,
                                       central_epoch_dec=Time(central_epoch_dec, format=format).value,
                                       central_epoch_ra=Time(central_epoch_ra, format=format).value,
                                       fit_degree=fit_degree,
                                       use_parallax=use_parallax,
                                       parallactic_pertubations=parallactic_pertubations,
                                       normed=normed)
        self.data = data
        self.fitter = fitter

    def fit(self, ra_vs_epoch, dec_vs_epoch, return_all=False):
        return self.fitter.fit_line(ra_vs_epoch=ra_vs_epoch, dec_vs_epoch=dec_vs_epoch, return_all=return_all)

    def optimal_central_epochs(self):
        return {'ra': self.fitter.find_optimal_central_epoch('ra'), 'dec': self.fitter.find_optimal_central_epoch('dec')}



"""
Module for fitting astrometric data.

Author: G. Mirek Brandt
"""
#from numba import jit
import numpy as np
import warnings
from htof.utils.fit_utils import ra_sol_vec, dec_sol_vec, chi2_matrix, transform_coefficients_to_unnormalized_domain
from htof.utils.fit_utils import chisq_of_fit


class AstrometricFitter(object):
    """
    :param inverse_covariance_matrices: ndarray of length epoch times with the 2x2 inverse covariance matrices
                                        for each epoch
    :param epoch_times: 1D array
                        array with each epoch. If these are years, then the returned proper motions will
                         be in Angle/year as well. The unit of epoch_times should be the same as central_epoch_ra
                         and central_epoch_dec.
    :param parallactic_pertubations: dict
           {'ra_plx': array-like, 'dec_plx': array-like}
           the pertubations from parallactic motion. e.g. central_ra + parallactic_pertubations['ra_plx']
           would give the skypath (in RA) of the object throughout the time observed, from parallax alone.
    :param fit_degree: int.
                       number of degrees in polynomial fit. fit_degree=1 with parallax would be a five parameter
                       fit. fit_degree=2 with parallax would be a 7 parameter fit.
    """
    def __init__(self, inverse_covariance_matrices=None, epoch_times=None,
                 astrometric_chi_squared_matrices=None, astrometric_solution_vector_components=None,
                 parallactic_pertubations=None, fit_degree=1, use_parallax=False,
                 central_epoch_ra=0, central_epoch_dec=0, normed=False):
        if normed:
            self._on_normed()
        if parallactic_pertubations is None:
            parallactic_pertubations = {'ra_plx': np.zeros_like(epoch_times),
                                        'dec_plx': np.zeros_like(epoch_times)}
        self.use_parallax = use_parallax
        self.parallactic_pertubations = parallactic_pertubations
        self.inverse_covariance_matrices = inverse_covariance_matrices
        self.epoch_times = epoch_times
        self.fit_degree = fit_degree
        self.central_epoch_dec = central_epoch_dec
        self.central_epoch_ra = central_epoch_ra
        self.normed = normed
        self.ra_epochs, self.dec_epochs = self._init_epochs()

        if astrometric_solution_vector_components is None:
            self.astrometric_solution_vector_components = self._init_astrometric_solution_vectors(fit_degree)
        if astrometric_chi_squared_matrices is None:
            self._chi2_matrix, astrometric_chi_squared_matrices = self._init_astrometric_chi_squared_matrix(fit_degree)
        self.astrometric_chi_squared_matrices = astrometric_chi_squared_matrices

    def find_optimal_central_epoch(self, coordinate='ra'):
        """
        Method to find the central_epoch_ra or central_epoch_dec (i.e. the reference epoch) that
        minimizes the covariance between
        the position and the proper motion (e.g. for 'ra' that minimizes the covariance between the ra offset
        and the proper motion in ra). Formally, the central_epoch is the epoch that produces 0 covariance between
        the proper motion and the position.
        From Brandt 2018 Eq 10 and the Hipparcos and Tycho Catalogues Vol X, Eq Y:
        for a covariance matrix evaluated at a reference epoch of `t`, then `\delta t` is
        `\delta t =` pm_pos_covariance / pm_variance
        where `t - \delta t =` the epoch where the covariance between proper motion and
        position (pm_pos_covariance) is zero.
        is zero.

        WARNING: including parallax or not in the fit changes the variance and covariance of the best fit
        parameters and therefore the central_epoch. For example, hip 27321 in Hipparcos 1 has central epochs
        as follows
        1991.2865 1991.246 # Ra, Dec, (with parallax)
        1991.2834 1991.254 # Ra, Dec, (without parallax)
        So including or excluding parallax changes the central epoch by up to ~ 0.01 years (in this case)
        or roughly 3 days. Some sources may have a larger shift, and so one should check this when considering
        whether or not they need parallax in their fit.

        :return: float. The reference epoch which makes the positional and proper motion covariance zero
        in the coordinate specified (either ra or dec).
        """
        if coordinate is not 'ra' and coordinate is not 'dec':
            raise ValueError('coordinate kwarg must be either ra or dec')
        # get the row (i) and column (j) that we are going to try and minimize in the covariance matrix
        row_coord, col_coord = {}, {}
        if coordinate is 'ra':
            row_coord['pm_pos_covariance'] = 1
            col_coord['pm_pos_covariance'] = 3
            row_coord['pm_variance'] = 3
            col_coord['pm_variance'] = 3
        if coordinate is 'dec':
            row_coord['pm_pos_covariance'] = 2
            col_coord['pm_pos_covariance'] = 4
            row_coord['pm_variance'] = 4
            col_coord['pm_variance'] = 4
        if not self.use_parallax:
            row_coord['pm_pos_covariance'] -= 1
            col_coord['pm_pos_covariance'] -= 1
            row_coord['pm_variance'] -= 1
            col_coord['pm_variance'] -= 1
        central_epoch = np.mean(self.epoch_times)
        for i in range(3):  # do a few iterations to converge to the correct central epoch value
            cov_matrix = self.evaluate_cov_matrix(central_epoch, central_epoch)
            pm_pos_covariance = cov_matrix[row_coord['pm_pos_covariance'], col_coord['pm_pos_covariance']]
            pm_variance = cov_matrix[row_coord['pm_variance'], col_coord['pm_variance']]
            central_epoch -= pm_pos_covariance / pm_variance
        return central_epoch

    def evaluate_cov_matrix(self, central_epoch_ra, central_epoch_dec):
        """
        Evaluate the covariance matrix of the fit at a particular set of central epochs. For use in
        find_optimal_central_epoch.
        :param central_epoch_ra:
        :param central_epoch_dec:
        :return:
        """
        icov_matrix = self._init_astrometric_chi_squared_matrix(self.fit_degree,
                                                                ra_epochs=np.array(self.epoch_times) - central_epoch_ra,
                                                                dec_epochs=np.array(self.epoch_times) - central_epoch_dec)[0]
        cov_matrix = np.linalg.pinv(icov_matrix, hermitian=True)
        return cov_matrix

    def _on_normed(self):
        warnings.warn('the normed fitting option (normed=True) will be removed in a future release.'
                      ' Do not use normed=True because it will cause the returned fit errors to be incorrect.',
                      PendingDeprecationWarning)

    def fit_line(self, ra_vs_epoch, dec_vs_epoch, return_all=False):
        """
        :param ra_vs_epoch: 1d array of right ascension, ordered the same as the covariance matrices and epochs.
        :param dec_vs_epoch: 1d array of declination, ordered the same as the covariance matrices and epochs.
        :param return_all: bool. True to return the solution vector as well as the 1-sigma error estimates on the parameters.
        :return: ndarray: best fit astrometric parameters.
                 E.g. [ra0, dec0, mu_ra, mu_dec] if use_parallax=False
                 or, [parallax_angle, ra0, dec0, mu_ra, mu_dec] if use_parallax=True
        """
        # performing the SVD fit.
        # linalg.pinv calculates the pseudo inverse via singular-value-decomposition.
        # hermitian=True forces the _chi2_matrix to be symmetric.
        cov_matrix = np.linalg.pinv(self._chi2_matrix, hermitian=True)
        solution = np.matmul(cov_matrix, self._chi2_vector(ra_vs_epoch=ra_vs_epoch, dec_vs_epoch=dec_vs_epoch))
        errors = np.sqrt(np.diagonal(cov_matrix))
        # calculating chisq of the fit.
        chisq = chisq_of_fit(solution, ra_vs_epoch, dec_vs_epoch,
                             self.ra_epochs, self.dec_epochs,
                             self.inverse_covariance_matrices, **self.parallactic_pertubations,
                             use_parallax=self.use_parallax)
        if self.normed:
            # transforming out of normalized coordinates.
            c_ra, c_dec = self.central_epoch_ra, self.central_epoch_dec
            t = self.epoch_times
            solution = transform_coefficients_to_unnormalized_domain(solution, t.min() - c_ra, t.max() - c_ra,
                                                                     t.min() - c_dec, t.max() - c_dec, self.use_parallax)
            # TODO : One should not transform the errors like the solution vector.
            # TODO: easy but bulky way: reconstruct the chi^2 matrix here.
            errors = transform_coefficients_to_unnormalized_domain(errors, t.min() - c_ra, t.max() - c_ra,
                                                                   t.min() - c_dec, t.max() - c_dec, self.use_parallax)
        return solution if not return_all else (solution, errors, chisq)

    def _chi2_vector(self, ra_vs_epoch, dec_vs_epoch):
        # this is the vector of partial derivatives of chisquared with respect to each astrometric parameter
        ra_solution_vecs = self.astrometric_solution_vector_components['ra']
        dec_solution_vecs = self.astrometric_solution_vector_components['dec']
        # sum together the individual solution vectors for each epoch
        return np.dot(ra_vs_epoch, ra_solution_vecs) + np.dot(dec_vs_epoch, dec_solution_vecs)

    def _init_astrometric_solution_vectors(self, fit_degree):
        num_epochs = len(self.epoch_times)
        plx = 1 * self.use_parallax
        astrometric_solution_vector_components = {'ra': np.zeros((num_epochs, 2 * fit_degree + 2 + plx)),
                                                  'dec': np.zeros((num_epochs, 2 * fit_degree + 2 + plx))}
        for obs in range(num_epochs):
            a, b, c, d = unpack_elements_of_matrix(self.inverse_covariance_matrices[obs])
            w_ra, w_dec = self.parallactic_pertubations['ra_plx'][obs], self.parallactic_pertubations['dec_plx'][obs]
            clip_i = 0 if self.use_parallax else 1
            astrometric_solution_vector_components['ra'][obs] = ra_sol_vec(a, b, c, d,
                                                                           self.ra_epochs[obs], self.dec_epochs[obs],
                                                                           w_ra, w_dec, deg=fit_degree)[clip_i:]
            astrometric_solution_vector_components['dec'][obs] = dec_sol_vec(a, b, c, d,
                                                                             self.ra_epochs[obs], self.dec_epochs[obs],
                                                                             w_ra, w_dec, deg=fit_degree)[clip_i:]
        return astrometric_solution_vector_components

    def _init_astrometric_chi_squared_matrix(self, fit_degree, ra_epochs=None, dec_epochs=None):
        if ra_epochs is None:
            ra_epochs = self.ra_epochs
        if dec_epochs is None:
            dec_epochs = self.dec_epochs
        num_epochs = len(self.epoch_times)
        plx = 1 * self.use_parallax
        astrometric_chi_squared_matrices = np.zeros((num_epochs, 2 * fit_degree + 2 + plx, 2 * fit_degree + 2 + plx))
        for obs in range(num_epochs):
            a, b, c, d = unpack_elements_of_matrix(self.inverse_covariance_matrices[obs])
            w_ra, w_dec = self.parallactic_pertubations['ra_plx'][obs], self.parallactic_pertubations['dec_plx'][obs]
            clip_i = 0 if self.use_parallax else 1
            astrometric_chi_squared_matrices[obs] = chi2_matrix(a, b, c, d,
                                                                ra_epochs[obs], dec_epochs[obs],
                                                                w_ra, w_dec, deg=fit_degree)[clip_i:, clip_i:]
        return np.sum(astrometric_chi_squared_matrices, axis=0), astrometric_chi_squared_matrices

    def _init_epochs(self):
        if not self.normed:
            # comment so that unit test registers.
            return np.array(self.epoch_times) - self.central_epoch_ra, np.array(self.epoch_times) - self.central_epoch_dec
        if self.normed:
            normed_epochs = normalize(self.epoch_times, [np.max(self.epoch_times), np.min(self.epoch_times)])
            return 1.*normed_epochs, 1.*normed_epochs


class AstrometricFastFitter(AstrometricFitter):
    """
    A faster version of AstrometricFitter. Can not return errors or the chisquared. Roughly 30 times faster
    per call of fit_line than AstrometricFitter.
    """
    def fit_line(self, ra_vs_epoch, dec_vs_epoch):
        """
        :param ra_vs_epoch: 1d array of right ascension, ordered the same as the covariance matrices and epochs.
        :param dec_vs_epoch: 1d array of declination, ordered the same as the covariance matrices and epochs.
        :return: ndarray: best fit astrometric parameters.
                 E.g. [ra0, dec0, mu_ra, mu_dec] if use_parallax=False
                 or, [parallax_angle, ra0, dec0, mu_ra, mu_dec] if use_parallax=True
        """
        return fast_fit_line(self._chi2_matrix, self.astrometric_solution_vector_components['ra'],
                             self.astrometric_solution_vector_components['dec'], ra_vs_epoch, dec_vs_epoch)

    def _on_normed(self):
        raise NotImplementedError('AstrometricFastFitter cannot implement the normed=True fit feature')


#@jit(nopython=True)
def fast_fit_line(chi2mat, ra_solution_vecs, dec_solution_vecs, ra_vs_epoch, dec_vs_epoch):
    return np.linalg.solve(chi2mat, np.dot(ra_vs_epoch, ra_solution_vecs) + np.dot(dec_vs_epoch, dec_solution_vecs))


def unpack_elements_of_matrix(matrix):
    return matrix.flatten()


def normalize(coordinates, domain):
    """
    :param coordinates: ndarray
    :param domain: ndarray. max and min value of input coordinates.
    :return: coordinates normalized to run from -1 to 1.
    """
    coordinates = 2. * (coordinates - min(domain))/(max(domain) - min(domain)) - 1.
    return coordinates



import numpy as np
from htof.polynomial import polynomial as poly


def test_taylorvander():
    expected = np.array([[1.,  1.,  0.5],
                         [1.,  2.,  2.],
                         [1.,  3.,  4.5],
                         [1.,  5., 12.5]])

    assert np.allclose(expected, poly.taylorvander(np.array([1, 2, 3, 5]), 2))


def test_taylor_series():
    series = poly.TaylorSeries(np.array([1, 2, 2, 1]))
    for domain in [np.arange(5), np.arange(5, 10), np.arange(-10, 5)]:
        assert np.allclose(series(domain), 1 + 2 * domain + 2 * 1/2 * domain**2 + 1/6 * domain**3)



"""
  Module for parsing intermediate data from Hipparcos and Gaia.
  For Hipparcos (both reductions) and Gaia, the scan angle theta is the angle between the north
  equitorial pole (declination) and the along-scan axis, defined as positive if east of the north pole
  (positive for increasing RA).

  Author:
    G. Mirek Brandt
    Daniel Michalik
"""

import numpy as np
import pandas as pd
import os
import re
import glob
import itertools
from math import ceil, floor
import pkg_resources

from astropy.time import Time
from astropy.table import QTable, Column, Table

from htof import settings as st
from htof.utils.data_utils import merge_consortia, safe_concatenate
from htof.utils.parse_utils import gaia_obmt_to_tcb_julian_year

import abc


class DataParser(object):
    """
    Base class for parsing Hip1, Hip2 and Gaia data. self.epoch, self.covariance_matrix and self.scan_angle are saved
    as pandas.DataFrame. use .values (e.g. self.epoch.values) to call the ndarray version.
    """
    def __init__(self, scan_angle=None, epoch=None, residuals=None, inverse_covariance_matrix=None,
                 along_scan_errs=None):
        self.scan_angle = pd.Series(scan_angle)
        self._epoch = pd.DataFrame(epoch)
        self.residuals = pd.Series(residuals)
        self.along_scan_errs = pd.Series(along_scan_errs)
        self.inverse_covariance_matrix = inverse_covariance_matrix
        self._rejected_epochs = []

    @staticmethod
    def read_intermediate_data_file(star_id: str, intermediate_data_directory: str, skiprows, header, sep):
        star_id = str(star_id)
        filepath = os.path.join(os.path.join(intermediate_data_directory, '**/'), '*' + star_id + '*')
        filepath_list = glob.glob(filepath, recursive=True)
        if len(filepath_list) != 1:
            # search for the star id with leading zeros stripped
            filepath = os.path.join(os.path.join(intermediate_data_directory, '**/'), '*' + star_id.lstrip('0') + '*')
            filepath_list = glob.glob(filepath, recursive=True)
        if len(filepath_list) != 1:
            # search for files with the full 6 digit hipparcos string
            filepath = os.path.join(os.path.join(intermediate_data_directory, '**/'), '*' + star_id.zfill(6) + '*')
            filepath_list = glob.glob(filepath, recursive=True)
        if len(filepath_list) != 1:
            # take the file with which contains only the hip id if there are multiple matches
            filepath = os.path.join(os.path.join(intermediate_data_directory, '**/'), '*' + star_id.lstrip('0') + '*')
            filepath_list = match_filename(glob.glob(filepath, recursive=True), star_id)
        if len(filepath_list) == 0:
            raise FileNotFoundError('No file with name containing {0} or {1} or {2} found in {3}'
                                    ''.format(star_id, star_id.lstrip('0'), star_id.zfill(6), intermediate_data_directory))
        if len(filepath_list) > 1:
            raise FileNotFoundError('Unable to find the correct file among the {0} files containing {1}'
                                    'found in {2}'.format(len(filepath_list), star_id, intermediate_data_directory))
        data = pd.read_csv(filepath_list[0], sep=sep, skiprows=skiprows, header=header, engine='python')
        return data

    @abc.abstractmethod
    def parse(self, star_id: str, intermediate_data_parent_directory: str, **kwargs):
        pass    # pragma: no cover

    def julian_day_epoch(self):
        return self._epoch.values.flatten()

    @property
    def epoch(self):
        return self._epoch.values.flatten()

    def calculate_inverse_covariance_matrices(self, cross_scan_along_scan_var_ratio=np.inf):
        self.inverse_covariance_matrix = calc_inverse_covariance_matrices(self.scan_angle,
                                                                          cross_scan_along_scan_var_ratio=cross_scan_along_scan_var_ratio,
                                                                          along_scan_errs=self.along_scan_errs)

    def write(self, path: str, *args, **kwargs):
        """
        :param path: str. filepath to write out the processed data.
        :param args: arguments for astropy.table.Table.write()
        :param kwargs: keyword arguments for astropy.table.Table.write()
        :return: None

        Note: The IntermediateDataParser.inverse_covariance_matrix are added to the table as strings
        so that they are easily writable. The icov matrix is saved a string.
        Each element of t['icov'] can be recovered with ast.literal_eval(t['icov'][i])
        where i is the index. ast.literal_eval(t['icov'][i]) will return a 2x2 list.
        """
        t = self.as_table()
        # transform icov matrices as writable strings.
        t['icov'] = [str(icov.tolist()) for icov in t['icov']]
        t.write(path, fast_writer=False, *args, **kwargs)

    def as_table(self):
        """
        :return: astropy.table.QTable
                 The IntermediateDataParser object tabulated.
                 This table has as columns all of the attributes of IntermediateDataParser.

                 For any attribute which is empty or None, the column will contain zeros.
        """
        cols = [self.scan_angle, self.julian_day_epoch(), self.residuals, self.along_scan_errs, self.inverse_covariance_matrix]
        cols = [Column(col) for col in cols]
        # replacing incorrect length columns with empties.
        cols = [col if len(col) == len(self) else Column(None, length=len(self)) for col in cols]

        t = QTable(cols, names=['scan_angle', 'julian_day_epoch', 'residuals', 'along_scan_errs', 'icov'])
        return t

    @property
    def rejected_epochs(self):
        return self._rejected_epochs

    @rejected_epochs.setter
    def rejected_epochs(self, value):
        not_outlier = np.ones(len(self), dtype=bool)
        np.put(not_outlier, value, False)
        self.along_scan_errs, self.scan_angle = self.along_scan_errs[not_outlier], self.scan_angle[not_outlier]
        self.residuals, self._epoch = self.residuals[not_outlier], self._epoch[not_outlier]
        self._rejected_epochs = value

    def __add__(self, other):
        all_scan_angles = pd.concat([self.scan_angle, other.scan_angle])
        all_epoch = pd.concat([pd.DataFrame(self.julian_day_epoch()), pd.DataFrame(other.julian_day_epoch())])
        all_residuals = pd.concat([self.residuals, other.residuals])
        all_along_scan_errs = pd.concat([self.along_scan_errs, other.along_scan_errs])

        all_inverse_covariance_matrix = safe_concatenate(self.inverse_covariance_matrix,
                                                         other.inverse_covariance_matrix)

        return DataParser(scan_angle=all_scan_angles, epoch=all_epoch, residuals=all_residuals,
                          inverse_covariance_matrix=all_inverse_covariance_matrix,
                          along_scan_errs=all_along_scan_errs)

    def __radd__(self, other):
        if other == 0:
            return self
        return self.__add__(other)

    def __len__(self):
        return len(self._epoch)


class GaiaData(DataParser):
    DEAD_TIME_TABLE_NAME = None

    def __init__(self, scan_angle=None, epoch=None, residuals=None, inverse_covariance_matrix=None,
                 min_epoch=-np.inf, max_epoch=np.inf, along_scan_errs=None):
        super(GaiaData, self).__init__(scan_angle=scan_angle, along_scan_errs=along_scan_errs,
                                       epoch=epoch, residuals=residuals,
                                       inverse_covariance_matrix=inverse_covariance_matrix)
        self.min_epoch = min_epoch
        self.max_epoch = max_epoch

    def parse(self, star_id, intermediate_data_directory, **kwargs):
        data = self.read_intermediate_data_file(star_id, intermediate_data_directory,
                                                skiprows=0, header='infer', sep='\s*,\s*')
        data = self.trim_data(data['ObservationTimeAtBarycentre[BarycentricJulianDateInTCB]'],
                              data, self.min_epoch, self.max_epoch)
        data = self.reject_dead_times(data['ObservationTimeAtBarycentre[BarycentricJulianDateInTCB]'], data)
        self._epoch = data['ObservationTimeAtBarycentre[BarycentricJulianDateInTCB]']
        self.scan_angle = data['scanAngle[rad]']

    def trim_data(self, epochs, data, min_mjd, max_mjd):
        valid = np.logical_and(epochs >= min_mjd, epochs <= max_mjd)
        return data[valid].dropna()

    def reject_dead_times(self, epochs, data):
        # there will be different astrometric gaps for gaia DR2 and DR3 because rejection criteria may change.
        # hence we have the appropriate parsers have different values for DEAD_TIME_TABLE_NAME.
        if self.DEAD_TIME_TABLE_NAME is None:
            # return the data if there is no dead time table specified.
            return data
        dead_time_table = Table.read(self.DEAD_TIME_TABLE_NAME)
        # convert on board mission time (OBMT) to julian day
        for col, newcol in zip(['start', 'end'], ['start_tcb_jd', 'end_tcb_jd']):
            dead_time_table[newcol] = gaia_obmt_to_tcb_julian_year(dead_time_table[col]).jd
        # make a mask of the epochs. Those that are within a dead time window have a value of 0 (masked)
        valid = np.ones(len(data), dtype=bool)
        for entry in dead_time_table:
            valid[np.logical_and(epochs >= entry['start_tcb_jd'], epochs <= entry['end_tcb_jd'])] = 0
        # reject the epochs which fall within a dead time window
        data = data[valid].dropna()
        return data


class DecimalYearData(DataParser):
    def __init__(self, scan_angle=None, epoch=None, residuals=None, inverse_covariance_matrix=None,
                 along_scan_errs=None):
        super(DecimalYearData, self).__init__(scan_angle=scan_angle, along_scan_errs=along_scan_errs,
                                              epoch=epoch, residuals=residuals,
                                              inverse_covariance_matrix=inverse_covariance_matrix)

    def parse(self, star_id, intermediate_data_parent_directory, **kwargs):
        pass  # pragma: no cover

    def julian_day_epoch(self):
        return Time(self._epoch.values.flatten(), format='decimalyear').jd


def calc_inverse_covariance_matrices(scan_angles, cross_scan_along_scan_var_ratio=np.inf,
                                     along_scan_errs=None):
    """
    :param scan_angles: pandas.DataFrame.
            data frame with scan angles, e.g. as-is from IntermediateDataParser.read_intermediate_data_file.
            scan_angles.values is a numpy array with the scan angles
    :param cross_scan_along_scan_var_ratio: var_cross_scan / var_along_scan
    :param along_scan_errs: array. array of len(scan_angles), the errors in the along scan direction, one for each
    scan in scan_angles.
    :return An ndarray with shape (len(scan_angles), 2, 2), e.g. an array of covariance matrices in the same order
    as the scan angles
    """
    if along_scan_errs is None or len(along_scan_errs) == 0:
        along_scan_errs = np.ones_like(scan_angles.values.flatten())
    icovariance_matrices = []
    icov_matrix_in_scan_basis = np.array([[1, 0],
                                         [0, 1/cross_scan_along_scan_var_ratio]])
    for theta, err in zip(scan_angles.values.flatten(), along_scan_errs):
        c, s = np.cos(theta), np.sin(theta)
        Rot = np.array([[s, -c], [c, s]])
        icov_matrix_in_ra_dec_basis = np.matmul(np.matmul(1/(err ** 2) * Rot, icov_matrix_in_scan_basis), Rot.T)
        icovariance_matrices.append(icov_matrix_in_ra_dec_basis)
    return np.array(icovariance_matrices)


class HipparcosOriginalData(DecimalYearData):
    def __init__(self, scan_angle=None, epoch=None, residuals=None, inverse_covariance_matrix=None,
                 along_scan_errs=None):
        super(HipparcosOriginalData, self).__init__(scan_angle=scan_angle, along_scan_errs=along_scan_errs,
                                                    epoch=epoch, residuals=residuals,
                                                    inverse_covariance_matrix=inverse_covariance_matrix)

    def parse(self, star_id, intermediate_data_directory, data_choice='MERGED'):
        """
        :param star_id: a string which is just the number for the HIP ID.
        :param intermediate_data_directory: the path (string) to the place where the intermediate data is stored, e.g.
                Hip2/IntermediateData/resrec
                note you have to specify the file resrec or absrec. We use the residual records, so specify resrec.
        :param data_choice: 'FAST' or 'NDAC', 'BOTH', or 'MERGED. The standard is 'MERGED' which does a merger
        of the 'NDAC' and 'FAST' data reductions in the same way as the hipparcos 1991.25 catalog. 'BOTH' keeps
        both consortia's data in the IAD, which would be unphysical and is just for debugging. 'FAST' would keep
        only the FAST consortia data, likewise only NDAC would be kept if you selected 'NDAC'.
        """
        if (data_choice is not 'NDAC') and (data_choice is not 'FAST') and (data_choice is not 'MERGED')\
                and (data_choice is not 'BOTH'):
            raise ValueError('data choice has to be either NDAC or FAST or MERGED or BOTH.')
        data = self.read_intermediate_data_file(star_id, intermediate_data_directory,
                                                skiprows=10, header='infer', sep='\s*\|\s*')
        data = self._fix_unnamed_column(data)
        data = self._select_data(data, data_choice)
        # compute scan angles and observations epochs according to van Leeuwen & Evans 1998
        #  10.1051/aas:1998218, eq. 11 & 12.
        self.scan_angle = np.arctan2(data['IA3'], data['IA4'])  # unit radians, arctan2(sin, cos)
        # Use the larger denominator when computing the epoch offset. 
        # This increases numerical precision and avoids NaNs if one of the two fields (IA3, IA4) is exactly zero.
        self._epoch = 1991.25 + (data['IA6'] / data['IA3']).where(abs(data['IA3']) > abs(data['IA4']), (data['IA7'] / data['IA4']))
        self.residuals = data['IA8']  # unit milli-arcseconds (mas)
        self.along_scan_errs = data['IA9']  # unit milli-arcseconds

    @staticmethod
    def _select_data(data, data_choice):
        # restrict intermediate data to either NDAC, FAST, or merge the NDAC and FAST results.
        if data_choice is 'MERGED':
            data = merge_consortia(data)
        elif data_choice is not 'BOTH':
            data = data[data['IA2'].str.upper() == {'NDAC': 'N', 'FAST': 'F'}[data_choice]]
        return data

    @staticmethod
    def _fix_unnamed_column(data, correct_key='IA2', col_idx=1):
        data.rename(columns={data.columns[col_idx]: correct_key}, inplace=True)
        return data


class HipparcosRereductionDVDBook(DecimalYearData):
    def __init__(self, scan_angle=None, epoch=None, residuals=None, inverse_covariance_matrix=None,
                 along_scan_errs=None):
        super(HipparcosRereductionDVDBook, self).__init__(scan_angle=scan_angle, along_scan_errs=along_scan_errs,
                                                          epoch=epoch, residuals=residuals,
                                                          inverse_covariance_matrix=inverse_covariance_matrix)

    def parse(self, star_id, intermediate_data_directory, error_inflate=True, header_rows=1, attempt_adhoc_rejection=True, **kwargs):
        """
        :param: star_id:
        :param: intermediate_data_directory:
        :param: error_inflate: True if the along-scan errors are to be corrected by the inflation factor
        according to equation B.1 of D. Michalik et al. 2014. Only turn this off for tests, or if the parameters
        required to compute the error inflation are unavailable.
        :param: header_rows: int.
        :return:

        Compute scan angles and observations epochs from van Leeuwen 2007, table G.8
        see also Figure 2.1, section 2.5.1, and section 4.1.2
        NOTE: that the Hipparcos re-reduction book and the figures therein describe the
        scan angle against the north ecliptic pole.
        NOTE: In the actual intermediate astrometry data on the CD the scan angle
        is given as east of the north equatorial pole, as for the original
        Hipparcos and Gaia (Source: private communication between Daniel
        Michalik and Floor van Leeuwen, April 2019).
        """
        header = self.read_intermediate_data_file(star_id, intermediate_data_directory,
                                                  skiprows=0, header=None, sep='\s+').iloc[0]
        data = self.read_intermediate_data_file(star_id, intermediate_data_directory,
                                                skiprows=header_rows, header=None, sep='\s+')
        self.scan_angle = np.arctan2(data[3], data[4])  # data[3] = sin(psi), data[4] = cos(psi)
        self._epoch = data[1] + 1991.25
        self.residuals = data[5]  # unit milli-arcseconds (mas)
        self.along_scan_errs = data[6]  # unit milli-arcseconds (mas)
        n_transits, nparam, catalog_f2, percent_rejected = header[2], get_nparam(header[4]), header[6], header[7]
        if attempt_adhoc_rejection:
            # must reject before inflating errors, otherwise F2 is around zero.
            epochs_to_reject = find_epochs_to_reject(self, catalog_f2, n_transits, nparam, percent_rejected)
            self.rejected_epochs = epochs_to_reject  # setting rejected_epochs also rejects the epochs (see the @setter)
        if error_inflate:
            # adjust the along scan errors so that the errors on the best fit parameters match the catalog.
            # TODO check that oyu sohuld incorporate n_reject as well.
            self.along_scan_errs *= self.error_inflation_factor(n_transits, nparam, catalog_f2)

    @staticmethod
    def error_inflation_factor(ntr, nparam, f2):
        """
        :param ntr: int. Number of transits used in the catalog solution. I.e. this should be
        N_transit_total - N_reject. So if N_reject is unknown, then the error inflation factor will be slightly wrong.
        :param nparam: int. Number of parameters used in the solution (e.g. 5, 7, 9..)
        :param f2: float. Goodness of fit metric. field F2 in the Hipparcos Re-reduction catalog.
        :return: u. float.
        The errors are to be scaled by u = Sqrt(Q/v) in equation B.4 of D. Michalik et al. 2014.
        (Title: Joint astrometric solution of Hipparcos and Gaia)
        NOTE: ntr (the number of transits) given in the header of the Hip2 IAD, is not necessarily
        the number of transits used.
        """
        num_transits_used = ntr
        nu = num_transits_used - nparam  # equation B.1 of D. Michalik et al. 2014
        Q = nu * (np.sqrt(2/(9*nu))*f2 + 1 - 2/(9*nu))**3  # equation B.3
        u = np.sqrt(Q/nu)  # equation B.4. This is the chi squared statistic of the fit.
        return u


class HipparcosRereductionJavaTool(HipparcosRereductionDVDBook):
    def __init__(self, scan_angle=None, epoch=None, residuals=None, inverse_covariance_matrix=None,
                 along_scan_errs=None):
        super(HipparcosRereductionJavaTool, self).__init__(scan_angle=scan_angle, along_scan_errs=along_scan_errs,
                                                         epoch=epoch, residuals=residuals,
                                                         inverse_covariance_matrix=inverse_covariance_matrix)

    def parse(self, star_id, intermediate_data_directory, **kwargs):
        # TODO set error error_inflate=True when the F2 value is available in the headers of 2.1 data.
        super(HipparcosRereductionJavaTool, self).parse(star_id, intermediate_data_directory,
                                                        error_inflate=False, header_rows=5, attempt_adhoc_rejection=False)
        epochs_to_reject = np.where(self.along_scan_errs < 0)[0]
        self.rejected_epochs = epochs_to_reject  # setting rejected_epochs also rejects the epochs (see the @setter)


class GaiaDR2(GaiaData):
    DEAD_TIME_TABLE_NAME = pkg_resources.resource_filename('htof', 'data/astrometric_gaps_gaiadr2_08252020.csv')

    def __init__(self, scan_angle=None, epoch=None, residuals=None, inverse_covariance_matrix=None,
                 min_epoch=st.GaiaDR2_min_epoch, max_epoch=st.GaiaDR2_max_epoch, along_scan_errs=None):
        super(GaiaDR2, self).__init__(scan_angle=scan_angle, along_scan_errs=along_scan_errs,
                                      epoch=epoch, residuals=residuals,
                                      inverse_covariance_matrix=inverse_covariance_matrix,
                                      min_epoch=min_epoch, max_epoch=max_epoch)


class GaiaeDR3(GaiaData):
    DEAD_TIME_TABLE_NAME = pkg_resources.resource_filename('htof', 'data/astrometric_gaps_gaiaedr3_12232020.csv')

    def __init__(self, scan_angle=None, epoch=None, residuals=None, inverse_covariance_matrix=None,
                 min_epoch=st.GaiaeDR3_min_epoch, max_epoch=st.GaiaeDR3_max_epoch, along_scan_errs=None):
        super(GaiaeDR3, self).__init__(scan_angle=scan_angle, along_scan_errs=along_scan_errs,
                                      epoch=epoch, residuals=residuals,
                                      inverse_covariance_matrix=inverse_covariance_matrix,
                                      min_epoch=min_epoch, max_epoch=max_epoch)


def digits_only(x: str):
    return re.sub("[^0-9]", "", x)


def match_filename(paths, star_id):
    return [f for f in paths if digits_only(os.path.basename(f).split('.')[0]).zfill(6) == star_id.zfill(6)]


def find_epochs_to_reject(data: DataParser, catalog_f2, n_transits, nparam, percent_rejected):
    atol_f2 = 0.1  # f2 must match to the catalog within this to be considered valid.
    # Calculate how many observations were probably rejected
    n_reject = max(floor((percent_rejected - 1) / 100 * n_transits), 0)
    max_n_reject = max(ceil((percent_rejected + 1) / 100 * n_transits), 1)
    # Calculate z_score and grab the worst max_n_reject observations plus a few for wiggle room.
    z_score = np.abs(data.residuals.values/data.along_scan_errs.values)
    possible_rejects = np.arange(len(data))[np.argsort(z_score)[::-1]][:max_n_reject + 3]
    # Calculate f2 without rejecting any observations
    chisquared = np.sum((data.residuals.values/data.along_scan_errs.values)**2)
    f2 = compute_f2(n_transits - nparam, chisquared)
    # Check if f2 agrees with the catalog.
    f2_matches_catalog = np.isclose(catalog_f2, f2, atol=atol_f2)
    #
    reject_idx = []
    if not f2_matches_catalog:
        # If f2 does not agree with the catalog, reject observations trying to match the f2 value.
        reject_idx_per_possibility = []
        f2_per_possibility = []
        while n_reject <= max_n_reject:
            # calculate f2 given sets of rejected observations of n_reject.
            n_reject += 1
            combinations = set(itertools.combinations(possible_rejects, n_reject))
            idx = np.ones((len(combinations), len(data)), dtype=bool)
            idx[[(i,) for i in range(len(combinations))], list(combinations)] = False
            chisquareds = np.nansum((data.residuals.values * idx / (data.along_scan_errs.values * idx))**2, axis=1)
            f2_trials = compute_f2(n_transits - n_reject - nparam, chisquareds)
            best_trial = np.nanargmin(np.abs(f2_trials - catalog_f2))
            f2_per_possibility.append(f2_trials[best_trial])
            reject_idx_per_possibility.append(list(list(combinations)[best_trial]))

        # see which reject combinations give an f2 value that is close to the catalog value (with wiggle room)
        reject_idx_viable = np.asarray(reject_idx_per_possibility)[np.isclose(catalog_f2, f2_per_possibility, atol=3 * atol_f2)]
        if len(reject_idx_viable) == 0:
            print('Could not find a set of epochs to reject that yielded an f2 value within {0} of the catalog value '
                  'for f2. Will proceed without rejecting ANY observations'.format(atol_f2))
            reject_idx = []
        elif len(reject_idx_viable) == 1:
            # if there is one viable combination, take that one
            reject_idx = reject_idx_viable[0]
        else:
            # if there are more than one viable combinations, take the combination that allows the catalog solution to be a stationary point.
            sum_squared_chisq_partials = []
            # evaluate the contributions to the different chisquared components
            sin_scan = np.sin(data.scan_angle.values)
            cos_scan = np.cos(data.scan_angle.values)
            dt = data.epoch - 1991.25
            chi2_vector = (2 * data.residuals.values / data.along_scan_errs.values ** 2 * np.array(
                [sin_scan, cos_scan, dt * sin_scan, dt * cos_scan])).T
            idx = np.ones(len(data), dtype=bool)
            for reject_idx in reject_idx_viable:
                # set the observations to be rejected.
                np.put(idx, reject_idx, False)
                # calculate the sum of the chisquared partials without the rejected observations
                sum_squared_chisq_partials.append(np.sum(np.sum(chi2_vector[idx], axis=0) ** 2))
                # reset the rejected observations for the next loop
                np.put(idx, reject_idx, True)
            reject_idx = reject_idx_viable[np.argmin(sum_squared_chisq_partials)]
        # Done trying to outlier reject.
        # Compute chisquared and f2 for final status check.
        idx = np.ones(len(data), dtype=bool)
        np.put(idx, reject_idx, False)
        chisquared = np.sum((data.residuals.values[idx] / data.along_scan_errs.values[idx]) ** 2)
        f2 = compute_f2(n_transits - nparam - len(reject_idx), chisquared)
        if not np.isclose(catalog_f2, f2, atol=0.05):
            print('catalog f2 value is {0} while the found value is {1}. It is possible that the '
                  'rejected observations numbered {2} are not the correct '
                  'rejections to make.'.format(catalog_f2, round(f2, 2), reject_idx))
    return reject_idx


def get_nparam(nparam_header_val):
    # strip the solution type (5, 7, or 9) from the solution type, which is a number 10xd+s consisting of
    # two parts: d and s. see Note 1 on Vizier for the Hipparcos re-reduction.
    return int(str(int(nparam_header_val))[-1])


def compute_f2(nu, chisquared):
    # equation B.2 of D. Michalik et al. 2014. Joint astrometric solution of Hipparcos and Gaia
    return (9*nu/2)**(1/2)*((chisquared/nu)**(1/3) + 2/(9*nu) - 1)



"""
NOTE 1: Gaia DR2 is based on data collected from the start of the nominal observations on 2014
 July 25 (10:30 UTC) until 2016 May 23 (11:35 UTC), or 668 days. However, the astrometric solution
 for this release did not use the observations during the first month after commissioning, when
 a special scanning mode (the ecliptic pole scanning law, EPSL) was employed. The data for the
 astrometry therefore start on 2014 Aug 22 21:00 UTC (BJD 2456892.375) and cover 640 days
 or 1.75 yr (therefore until 2016 May 23, or BJD 2457532.375).


NOTE 2: GaiaEDR3 is based on data collected from the start of the nom-inal
observations  on  25  July  2014  (10:30  UTC)  until  28  May2017 (08:45 UTC),
or 1038 days (data segments DS0–DS3 inFig. 1). Similarly to the astrometric solution for
DR2 (Lindegrenet al. 2018), this solution did not use the observations in the first month of
the operational phase, when the special ecliptic polescanning law (EPSL) was employed.
The data for the astrom-etry therefore start on 22 August 2014 (21:00 UTC) and cover
1009 days or 2.76 yr, with some interruptions mentioned below.

NOTE 2 is from section 2.2 of https://www.aanda.org/articles/aa/pdf/forth/aa39709-20.pdf
"""

GaiaDR2_min_epoch = 2456892.375  # Barycentric Julian Day (BJD)
GaiaDR2_max_epoch = 2457532.375  # Barycentric Julian Day (BJD)

GaiaeDR3_min_epoch = 2456892.375  # Barycentric Julian Day (BJD), 2014.6403 in jyear
GaiaeDR3_max_epoch = 2457901.375  # Barycentric Julian Day (BJD), 2017.4028 in jyear



from htof.validation.utils import refit_hip1_object, refit_hip2_object, load_hip2_catalog, refit_hip21_object
from htof.validation.utils import load_hip1_dm_annex, load_hip2_seven_p_annex, load_hip2_nine_p_annex
import os
from astropy.table import Table
from argparse import ArgumentParser
from glob import glob

from multiprocessing import Pool


class Engine(object):
    @staticmethod
    def format_result(result, hip_id, soltype):
        diffs, errors, chisq, partials = result[:4]
        plx, ra, dec, pm_ra, pm_dec, acc_ra, acc_dec, jerk_ra, jerk_dec = diffs
        return {'hip_id': hip_id, 'diff_ra': ra, 'diff_dec': dec, 'diff_plx': plx, 'diff_pm_ra': pm_ra, 'diff_pm_dec': pm_dec,
                'soltype': soltype, 'diff_acc_ra': acc_ra, 'diff_acc_dec': acc_dec, 'diff_jerk_ra': jerk_ra, 'diff_jerk_dec': jerk_dec,
                'chisquared': chisq, 'dxdra0': partials[0], 'dxddec0': partials[1], 'dxdmura': partials[2], 'dxdmudec': partials[3]}


class Hip1Engine(Engine):
    def __init__(self, dirname, use_parallax, hip_dm_g=None):
        self.dirname = dirname
        self.use_parallax = use_parallax
        self.hip_dm_g = hip_dm_g

    def __call__(self, fname):
        hip_id = os.path.basename(fname).split('.txt')[0]
        result = refit_hip1_object(self.dirname, hip_id, self.hip_dm_g, use_parallax=self.use_parallax)
        soltype = result[4]
        return self.format_result(result, hip_id, soltype)


class Hip2Engine(Engine):
    def __init__(self, dirname, use_parallax, catalog=None, seven_p_annex=None, nine_p_annex=None):
        self.dirname = dirname
        self.catalog = catalog
        self.use_parallax = use_parallax
        self.seven_p_annex = seven_p_annex
        self.nine_p_annex = nine_p_annex

    def __call__(self, fname):
        hip_id = os.path.basename(fname).split('.d')[0].split('HIP')[1]
        result = refit_hip2_object(self.dirname, hip_id, self.catalog, seven_p_annex=self.seven_p_annex,
                                   nine_p_annex=self.nine_p_annex, use_parallax=self.use_parallax)
        soltype = result[4]
        return self.format_result(result, hip_id, soltype[-1])


class Hip21Engine(Engine):
    def __init__(self, dirname, use_parallax):
        self.dirname = dirname
        self.use_parallax = use_parallax

    def __call__(self, fname):
        hip_id = os.path.basename(fname).split('.csv')[0].split('H')[1]
        result = refit_hip21_object(self.dirname, hip_id, use_parallax=self.use_parallax)
        soltype = result[4]
        return self.format_result(result, hip_id, soltype[-1])


if __name__ == "__main__":
    parser = ArgumentParser(description='Script for refitting the entire hipparcos catalog, 1997 or 2007.'
                                        'This will output a csv type file. Each row gives'
                                        'the difference in the best-fit parameters and the catalog parameters '
                                        'along with other metrics of interest.')
    parser.add_argument("-dir", "--iad-directory", required=True, default=None,
                        help="full path to the intermediate data directory")
    parser.add_argument("-hr", "--hip-reduction", required=True, default=None, type=int,
                        help="integer. 1 for 1997 reduction, 2 for 2007 CD reduction, 21 for 2007 IADT tool.")
    parser.add_argument("-o", "--output-file", required=False, default=None,
                        help="The output filename, with .csv extension. E.g. hip1_refit.csv."
                             "Will default to hip_processid.csv.")
    parser.add_argument("-c", "--cores", required=False, default=1, type=int,
                        help="Number of cores to use. Default is 1.")
    parser.add_argument("--ignore-parallax", required=False, action='store_true', default=False,
                        help="Whether or not to ignore parallax in the fits. Default is False, i.e. parallax"
                             "will be fit.")
    parser.add_argument("-cpath", "--catalog-path", required=False, default=None,
                        help="path to the Hip re-reduction main catalog, e.g. Main_cat.d. Only required"
                             "if using the 2007 CD data.")
    parser.add_argument("--debug", action='store_true', default=False, required=False,
                        help='If true, this will run the refit test on only 500 sources. Useful to check for '
                             'filepath problems before running the full test on all ~100000 sources.')

    args = parser.parse_args()

    # check arguments and assign values.
    if args.hip_reduction == 2 and args.catalog_path is None:
        raise ValueError('Hip 2 selected but no --catalog-path provided.')

    if args.output_file is None:
        output_file = 'hip' + args.hip_reduction + '_refit' + (str)(os.getpid()) + '.csv'
    else:
        output_file = args.output_file
    if args.ignore_parallax:
        print('Warning: ignore_parallax flag is active. parallax will not be fit and so the standard errors will '
              'be slightly different compared to the catalog values for every source')

    # find the intermediate data files
    kwargs = {}
    if args.hip_reduction == 1:
        files = glob(os.path.join(args.iad_directory, '*.txt'))
        engine = Hip1Engine
        kwargs = {'hip_dm_g': load_hip1_dm_annex(os.path.join(args.iad_directory, 'hip_dm_g.dat'))}
    elif args.hip_reduction == 2:
        files = glob(os.path.join(args.iad_directory, '**/H*.d'))
        engine = Hip2Engine
        ninep_path = os.path.join(os.path.dirname(args.catalog_path), 'NineP_Cat.d')
        sevenp_path = os.path.join(os.path.dirname(args.catalog_path), 'SevenP_Cat.d')
        kwargs = {'catalog': load_hip2_catalog(args.catalog_path),
                  'seven_p_annex': load_hip2_seven_p_annex(sevenp_path), 'nine_p_annex': load_hip2_nine_p_annex(ninep_path)}
    else:
        files = glob(os.path.join(args.iad_directory, '**/H*.csv'))
        engine = Hip21Engine
    # fit a small subset of sources if debugging.
    if args.debug:
        files = files[:500]
    print('will fit {0} total hip {1} objects'.format(len(files), str(args.hip_reduction)))
    print('will save output table at', output_file)
    # do the fit.
    try:
        pool = Pool(args.cores)
        engine = engine(args.iad_directory, not args.ignore_parallax, **kwargs)
        data_outputs = pool.map(engine, files)
        out = Table(data_outputs)
        out.sort('hip_id')
        out.write(output_file, overwrite=True)
    finally:  # This makes sure processes are closed in the end, even if errors happen
        pool.close()
        pool.join()




from astropy.time import Time


def gaia_obmt_to_tcb_julian_year(obmt):
    """
    convert OBMT (on board mission timeline) to TCB julian years via
    https://gea.esac.esa.int/archive/documentation/GDR2/Introduction/chap_cu0int/cu0int_sec_release_framework/cu0int_ssec_time_coverage.html
    Equation 1.1
    :param obmt: on-board mission timeline in units of six-hour revolutions since launch. OBMT (in revolutions)
    :return: astropy.time.Time

    Note that this is the same for DR2 as it is for eDR3, as of 12 23 2020.
    """
    tcbjy = 2015 + (obmt - 1717.6256)/(1461)
    return Time(tcbjy, scale='tcb', format='jyear')



"""
This module provides functions for calculating the astrometric paths of stars on the sky as seen by an
observer in orbit around the solar system barycentre. That is, the topocentric coordinate directions as a
function of time are calculated.

Author: Anthony Brown Nov 2018 - Nov 2018

From https://github.com/agabrown/astrometric-sky-path/  from commit: 039768eae04dca0b9b6615cccfc021e6a381bf4d
Reproduced with permission of the author.
"""

import numpy as np

from astropy import constants
from astropy import units
from astropy.time import Time
from astropy.coordinates import Angle
from astropy.coordinates import get_body_barycentric


def earth_ephemeris(t):
    """
    Calculate the ephemeris for the earth in the BCRS using astropy tools.
    
    NOTE: There are several versions of the solar system ephemeris available in astropy and others can be
    provided through a URL (see the documentation of astropy.coordinates.solar_system_ephemeris).
    Depending on the accuracy needed, employing  an ephemeris different from the default may be better.

    Parameters
    ----------
    
    t : array
        Barycentric Julian Year times at which to calculate the ephemeris.
        
    Returns
    -------

    Array of shape (3,t.size) representing the xyz components of the ephemeris at times t.

    Note: the units of angle should be radians and the units of time Julian years ('jyear'),
    while the distance unit is the AU.
    """
    times = Time(t, format='jyear', scale='tcb')  # 'tcb': Barycentric Coordinate Time (TCB)
    ephemeris = get_body_barycentric('earth', times)  # unit A.U.
    return np.vstack((ephemeris.x.value, ephemeris.y.value, ephemeris.z.value))


def earth_sun_l2_ephemeris(t):
    """
    Calculate the ephemeris for earth-sun L2 point in the BCRS using astropy tools.

    :param t: float array.
    Times at which to calculate the ephemeris in Julian years TCB.
    :return: float array.
    Array of shape (3,t.size) representing the xyz components of the ephemeris at times t.

    Note: 1.511 / 1.496 is the ratio of L2 semi-major axis to the earth's semi-major axis.
    """
    return earth_ephemeris(t) * 1.511 / 1.496


def epoch_topocentric_coordinates(alpha, delta, parallax, mura, mudec, vrad, t, refepoch, ephem):
    """
    For each observation epoch calculate the topocentric (as seen from a location on the earth's surface)
    coordinate directions (alpha(t), delta(t)) given the astrometric parameters of a source, the observation times,
    and the ephemeris (in the BCRS) for the observer. Also calculate the local plane coordinates xi(t) and eta(t).

    The code is partly based on the SOFA library (http://www.iausofa.org/) pmpx.c code.

    Parameters
    ----------
    
    alpha : float
        Right ascension at reference epoch (radians)
    delta : float
        Declination at reference epoch (radians)
    parallax : float
        Parallax (mas), negative values allowed
    mura : float
        Proper motion in right ascension, including cos(delta) factor (mas/yr)
    mudec : float
        Proper motion in declination (mas/yr)
    vrad : float
        Radial velocity (km/s)
    t : float array
        Observation times (Julian year TCB)
    refepoch : float
        Reference epoch (Julian year TCB)
    ephem : function
        Function providing the observer's ephemeris in BCRS at times t (units of AU)
                
    Returns (array, array, array, array)
    -------
    
    Arrays alpha, delta, xi, eta. Units are radians for (alpha, delta) and rad for (xi, eta).
    """
    # unit conversions
    _radtomas = (180 * 3600 * 1000) / np.pi
    _mastorad = np.pi / (180 * 3600 * 1000)
    _kmps_to_aupyr = (units.year.to(units.s) * units.km.to(units.m)) / constants.au.value

    # Normal triad, defined at the reference epoch.
    p = np.array([-np.sin(alpha), np.cos(alpha), 0.0])
    q = np.array([-np.sin(delta) * np.cos(alpha), -np.sin(delta) * np.sin(alpha), np.cos(delta)])
    r = np.array([np.cos(delta) * np.cos(alpha), np.cos(delta) * np.sin(alpha), np.sin(delta)])
   
    # Calculate observer's ephemeris.
    bO_bcrs = ephem(t)

    # Calculate the Roemer delay, take units into account.
    tB = t + np.dot(r, bO_bcrs) * constants.au.value / constants.c.value / units.year.to(units.s)
    
    plxrad = parallax*_mastorad
    murarad = mura*_mastorad
    mudecrad = mudec*_mastorad
    mur = vrad*_kmps_to_aupyr*np.abs(plxrad)

    uO = np.repeat(r, t.size).reshape((r.size, t.size))
    uO = uO + np.tensordot((p*murarad + q*mudecrad + r*mur), (tB-refepoch), axes=0) - plxrad*bO_bcrs
    
    # Local plane coordinates which approximately equal delta_alpha*cos(delta) and delta_delta
    xi = np.dot(p, uO)/np.dot(r, uO)
    eta = np.dot(q, uO)/np.dot(r, uO)

    alpha_obs = np.arctan2(uO[1, :], uO[0, :])
    delta_obs = np.arctan2(uO[2, :], np.sqrt(uO[0, :]**2+uO[1, :]**2))
                 
    return alpha_obs, delta_obs, xi, eta


def parallactic_motion(epochs, cntr_ra, cntr_dec, unit, refepoch, ephemeris=earth_ephemeris, parallax=1):
    """
    :param epochs: array of times in Julian year. Use astropy.time.Time.jyear to convert outside of this.
    :param cntr_ra: right ascension coordinate about which to calculate the parallactic motion. Should be in
                    the appropriate form for having units of unit.
    :param cntr_dec: declination coordinate about which to calculate the parallactic motion. Should be in
                    the appropriate form for having units of unit.
    :param unit: from Astropy.unit. Must be such that astropy.coordinates.Angle(cntr_ra, unit=unit) is sensical.
    :param refepoch: reference epoch in julian year.
    :param parallax: float. The parallax angle in milli-arcseconds.
    :param ephemeris: function.
          Function which intakes an array of Julian years and returns an array of shape (3,t.size)
          with the xyz components of the ephemeris at times t (along rows 0, 1 and 2 respectively).
    :return: [array, array]
    parallax motion about the center coordinate. E.g. Parallax_ra - cntr_ra and Parallax_dec - cntr_dec
    Where Parallax_ra would be an array of RA coordinates for parallax motion alone
    Output will have units of unit.
    """
    delta_ra, delta_dec = epoch_topocentric_coordinates(Angle(cntr_ra, unit=unit).rad,
                                                    Angle(cntr_dec, unit=unit).rad, parallax,
                                                    mura=0, mudec=0, vrad=0, t=epochs,
                                                    refepoch=refepoch, ephem=ephemeris)[2:]
    delta_ra, delta_dec = Angle(delta_ra, unit='radian').to(unit).value, Angle(delta_dec, unit='radian').to(unit).value
    return delta_ra, delta_dec


import pytest
import os
import numpy as np
from astropy.coordinates import Angle
from astropy.time import Time

from htof.parse import GaiaeDR3, GaiaData, HipparcosRereductionDVDBook, HipparcosOriginalData
from htof.fit import AstrometricFitter
from htof.validation.utils import refit_hip2_object, refit_hip1_object, load_hip2_catalog, load_hip2_seven_p_annex, load_hip2_nine_p_annex
from htof.validation.utils import load_hip1_dm_annex

from htof.main import Astrometry


@pytest.mark.integration
def test_parse_and_fit_to_line():
    """
    Tests fitting a line to fake RA and DEC data which has errors calculated from the real intermediate data
    from Hip1, Hip2, and GaiaDR2. This only fits a line to the first 11 points.
    """
    stars = ['049699', '027321', '027321']
    parsers = [GaiaeDR3, GaiaData, HipparcosOriginalData, HipparcosRereductionDVDBook]
    subdirectories = ['GaiaeDR3', 'GaiaDR2', 'Hip1', 'Hip2']
    base_directory = os.path.join(os.getcwd(), 'htof/test/data_for_tests')
    for star, parser, subdirectory in zip(stars, parsers, subdirectories):
        test_data_directory = os.path.join(base_directory, subdirectory)
        data = parser()
        data.parse(star_id=star,
                   intermediate_data_directory=test_data_directory)
        data.calculate_inverse_covariance_matrices()
        fitter = AstrometricFitter(inverse_covariance_matrices=data.inverse_covariance_matrix,
                                   epoch_times=np.linspace(0, 10, num=11))
        solution_vector = fitter.fit_line(ra_vs_epoch=np.linspace(30, 40, num=11),
                                          dec_vs_epoch=np.linspace(20, 30, num=11))
        ra0, dec0, mu_ra, mu_dec = solution_vector

        assert np.isclose(ra0, 30)
        assert np.isclose(dec0, 20)
        assert np.isclose(mu_ra, 1)
        assert np.isclose(mu_dec, 1)


class TestHipReReductionCDFits:
    CATALOG = load_hip2_catalog('htof/test/data_for_tests/Hip2/truncated_hip2dvd_Main_Cat.d')
    NINEP = load_hip2_nine_p_annex('htof/test/data_for_tests/Hip2/NineP_Cat.d')
    SEVENP = load_hip2_seven_p_annex('htof/test/data_for_tests/Hip2/SevenP_Cat.d')

    @pytest.mark.e2e
    @pytest.mark.parametrize("hip_id", ['70', '78999', '27321'])
    def test_Hip2_fit_5p_source(self, hip_id):
        diffs, error_diffs, chisq, chi2_partials, soltype = refit_hip2_object('htof/test/data_for_tests/Hip2',
                                                                         hip_id, catalog=self.CATALOG, use_parallax=True)
        assert np.allclose(diffs, 0, atol=0.02)
        assert np.allclose(error_diffs, 0, atol=0.1)

    @pytest.mark.e2e
    @pytest.mark.parametrize("hip_id", ['9631', '16468', '25838'])
    def test_Hip2_fit_7p9p_source(self, hip_id):
        diffs, error_diffs, chisq, chi2_partials, soltype = refit_hip2_object('htof/test/data_for_tests/Hip2/IntermediateData', hip_id,
                                                                              nine_p_annex=self.NINEP, seven_p_annex=self.SEVENP,
                                                                              catalog=self.CATALOG, use_parallax=True)
        assert np.allclose(diffs, 0, atol=0.02)
        # TODO test the error differences!
        #assert np.allclose(error_diffs, 0, atol=0.1)


class TestHip1Fits:
    SEVEN_NINEP_ANNEX = load_hip1_dm_annex('htof/test/data_for_tests/Hip1/hip_dm_g.dat')

    @pytest.mark.e2e
    @pytest.mark.parametrize("hip_id", ['5310', '5313', '46871', '50103', '46979'])
    def test_Hip1_fit_to_hip7p9p_source(self, hip_id):
        diffs, errors, chisq, chi2_partials, soltype = refit_hip1_object('htof/test/data_for_tests/Hip1/IntermediateData', hip_id,
                                                                         hip_dm_g=self.SEVEN_NINEP_ANNEX,
                                                                         use_parallax=True)
        assert np.allclose(diffs, 0, atol=0.07)

    @pytest.mark.e2e
    @pytest.mark.parametrize("hip_id", ['027321', '004391', '044801', '70000'])
    def test_Hip1_fit(self, hip_id):
        diffs, errors, chisq, chi2_partials, soltype = refit_hip1_object('htof/test/data_for_tests/Hip1/IntermediateData', hip_id,
                                                                         hip_dm_g=self.SEVEN_NINEP_ANNEX,
                                                                         use_parallax=True)
        assert np.allclose(diffs, 0, atol=0.02)


@pytest.mark.e2e
def test_Hip1_fit_to_hip27321():
    # Hip 27321 parameters from the Hipparcos 1 catalogue via Vizier
    cntr_ra, cntr_dec = Angle(86.82118054, 'degree'), Angle(-51.06671341, 'degree')
    plx = 51.87  # mas
    pmRA = 4.65  # mas/year
    pmDec = 81.96  # mas/year
    # generate fitter and parse intermediate data
    astro = Astrometry('Hip1', '27321', 'htof/test/data_for_tests/Hip1', central_epoch_ra=1991.25,
                       central_epoch_dec=1991.25, format='jyear', fit_degree=1, use_parallax=True,
                       central_ra=cntr_ra, central_dec=cntr_dec, normed=False)
    chisq = np.sum(astro.data.residuals ** 2 / astro.data.along_scan_errs ** 2)
    # generate ra and dec for each observation.
    year_epochs = Time(astro.data.julian_day_epoch(), format='jd', scale='tcb').jyear - \
                  Time(1991.25, format='decimalyear').jyear
    ra_motion = astro.fitter.parallactic_pertubations['ra_plx']
    dec_motion = astro.fitter.parallactic_pertubations['dec_plx']
    ra = Angle(ra_motion * plx + pmRA * year_epochs, unit='mas')
    dec = Angle(dec_motion * plx + pmDec * year_epochs, unit='mas')
    # add residuals
    ra += Angle(astro.data.residuals.values * np.sin(astro.data.scan_angle.values), unit='mas')
    dec += Angle(astro.data.residuals.values * np.cos(astro.data.scan_angle.values), unit='mas')
    #
    coeffs, errors, chisq_found = astro.fit(ra.mas, dec.mas, return_all=True)
    assert np.isclose(chisq, chisq_found, atol=1E-3)
    assert np.allclose([pmRA, pmDec], np.array([coeffs[3], coeffs[4]]).round(2))
    assert np.isclose(plx, coeffs[0].round(2), atol=0.01)
    assert np.allclose(errors.round(2), np.array([0.51, 0.45, 0.46, 0.53, 0.61]))


@pytest.mark.e2e
def test_Hip1_fit_to_hip44801():
    # Hip 44801 parameters from the intermediate data
    cntr_ra, cntr_dec = Angle(136.94995265, 'degree'), Angle(-9.85368471, 'degree')
    plx = 2.90  # mas
    pmRA = -10.91  # mas/year
    pmDec = 5.06  # mas/year
    # generate fitter and parse intermediate data
    astro = Astrometry('Hip1', '44801', 'htof/test/data_for_tests/Hip1', central_epoch_ra=1991.25,
                       central_epoch_dec=1991.25, format='jyear', fit_degree=1, use_parallax=True,
                       central_ra=cntr_ra, central_dec=cntr_dec, normed=False)
    chisq = np.sum(astro.data.residuals ** 2 / astro.data.along_scan_errs ** 2)
    # generate ra and dec for each observation.
    year_epochs = Time(astro.data.julian_day_epoch(), format='jd', scale='tcb').jyear - \
                  Time(1991.25, format='decimalyear').jyear
    ra_motion = astro.fitter.parallactic_pertubations['ra_plx']
    dec_motion = astro.fitter.parallactic_pertubations['dec_plx']
    ra = Angle(ra_motion * plx + pmRA * year_epochs, unit='mas')
    dec = Angle(dec_motion * plx + pmDec * year_epochs, unit='mas')
    # add residuals
    ra += Angle(astro.data.residuals.values * np.sin(astro.data.scan_angle.values), unit='mas')
    dec += Angle(astro.data.residuals.values * np.cos(astro.data.scan_angle.values), unit='mas')
    #
    coeffs, errors, chisq_found = astro.fit(ra.mas, dec.mas, return_all=True)
    assert np.isclose(chisq, chisq_found, atol=1E-3)
    assert np.allclose([pmRA, pmDec], np.array([coeffs[3], coeffs[4]]).round(2))
    assert np.isclose(plx, coeffs[0].round(2), atol=0.01)
    assert np.allclose(errors.round(2), np.array([1.09, 0.88, 0.77, 1.05, 0.80]))


@pytest.mark.e2e
def test_Hip1_fit_to_hip70000():
    # Hip 70000 parameters from the intermediate data
    cntr_ra, cntr_dec = Angle(214.85975459, 'degree'), Angle(14.93570946, 'degree')
    plx = 1.26  # mas
    pmRA = 1.01  # mas/year
    pmDec = 7.27  # mas/year
    # generate fitter and parse intermediate data
    astro = Astrometry('Hip1', '70000', 'htof/test/data_for_tests/Hip1', central_epoch_ra=1991.25,
                       central_epoch_dec=1991.25, format='jyear', fit_degree=1, use_parallax=True,
                       central_ra=cntr_ra, central_dec=cntr_dec, normed=False)
    chisq = np.sum(astro.data.residuals ** 2 / astro.data.along_scan_errs ** 2)
    # generate ra and dec for each observation.
    year_epochs = Time(astro.data.julian_day_epoch(), format='jd', scale='tcb').jyear - \
                  Time(1991.25, format='decimalyear').jyear
    ra_motion = astro.fitter.parallactic_pertubations['ra_plx']
    dec_motion = astro.fitter.parallactic_pertubations['dec_plx']
    ra = Angle(ra_motion * plx + pmRA * year_epochs, unit='mas')
    dec = Angle(dec_motion * plx + pmDec * year_epochs, unit='mas')
    # add residuals
    ra += Angle(astro.data.residuals.values * np.sin(astro.data.scan_angle.values), unit='mas')
    dec += Angle(astro.data.residuals.values * np.cos(astro.data.scan_angle.values), unit='mas')
    #
    coeffs, errors, chisq_found = astro.fit(ra.mas, dec.mas, return_all=True)
    assert np.isclose(chisq, chisq_found, atol=1E-3)
    assert np.allclose([pmRA, pmDec], np.array([coeffs[3], coeffs[4]]).round(2))
    assert np.isclose(plx, coeffs[0].round(2), atol=0.01)
    assert np.allclose(errors.round(2), np.array([1.11, 0.79, 0.62, 0.82, 0.64]))


@pytest.mark.e2e
def test_Hip1_fit_to_hip27321_no_parallax():
    # WARNING including the parallax component is important if you want to recover the catalog errors.
    # Hip 27321 parameters from the Hipparcos 1 catalogue via Vizier
    pmRA = 4.65  # mas/year
    pmDec = 81.96  # mas/year
    # generate fitter and parse intermediate data
    astro = Astrometry('Hip1', '27321', 'htof/test/data_for_tests/Hip1', central_epoch_ra=1991.25,
                       central_epoch_dec=1991.25, format='jyear', fit_degree=1,
                       use_parallax=False, normed=False)
    chisq = np.sum(astro.data.residuals ** 2 / astro.data.along_scan_errs ** 2)
    # generate ra and dec for each observation.
    year_epochs = Time(astro.data.julian_day_epoch(), format='jd', scale='tcb').jyear - \
                  Time(1991.25, format='decimalyear').jyear
    ra = Angle(pmRA * year_epochs, unit='mas')
    dec = Angle(pmDec * year_epochs, unit='mas')
    # add residuals
    ra += Angle(astro.data.residuals.values * np.sin(astro.data.scan_angle.values), unit='mas')
    dec += Angle(astro.data.residuals.values * np.cos(astro.data.scan_angle.values), unit='mas')
    #
    coeffs, errors, chisq_found = astro.fit(ra.mas, dec.mas, return_all=True)
    assert np.isclose(chisq, chisq_found, atol=1E-3)
    assert np.allclose([pmRA, pmDec], np.array([coeffs[2], coeffs[3]]).round(2))
    assert np.allclose(errors.round(2), np.array([0.45, 0.46, 0.53, 0.61]), atol=0.01)


@pytest.mark.e2e
def test_optimal_central_epochs_forHip1_hip27321():
    # Hip 27321 parameters from the Hipparcos 1 catalogue via Vizier
    cntr_ra, cntr_dec = Angle(86.82118054, 'degree'), Angle(-51.06671341, 'degree')
    # generate fitter and parse intermediate data
    astro = Astrometry('Hip1', '27321', 'htof/test/data_for_tests/Hip1', central_epoch_ra=1991.25,
                       central_epoch_dec=1991.25, format='jyear', fit_degree=1, use_parallax=True,
                       central_ra=cntr_ra, central_dec=cntr_dec, normed=False)
    central_epoch = astro.optimal_central_epochs()
    central_epoch_ra, central_epoch_dec = central_epoch['ra'], central_epoch['dec']
    #print(central_epoch_ra, central_epoch_dec)
    fitter = astro.fitter
    cov_matrix = fitter.evaluate_cov_matrix(central_epoch_ra, central_epoch_dec)
    ra_mura_cov, dec_mudec_cov = cov_matrix[1, 3], cov_matrix[2, 4]
    # do a brute force evaluation of all reasonable central epochs
    epoch_t = np.linspace(1991, 1992, 200)
    ra_vals, dec_vals = [], []
    for t in epoch_t:
        cov = fitter.evaluate_cov_matrix(t, t)
        ra_vals.append(cov[1, 3])
        dec_vals.append(cov[2, 4])
    # assert that the optimal central epochs give better covariances than all of those.
    assert np.all(np.abs(dec_vals) >= dec_mudec_cov)
    assert np.all(np.abs(ra_vals) >= ra_mura_cov)
    assert np.allclose([dec_mudec_cov, ra_mura_cov], 0, atol=1e-8)


@pytest.mark.e2e
def test_optimal_central_epochs_forHip1_hip27321_no_parallax():
    astro = Astrometry('Hip1', '27321', 'htof/test/data_for_tests/Hip1', central_epoch_ra=1991.25,
                       central_epoch_dec=1991.25, format='jyear', fit_degree=1,
                       use_parallax=False, normed=False)
    central_epoch = astro.optimal_central_epochs()
    central_epoch_ra, central_epoch_dec = central_epoch['ra'], central_epoch['dec']
    #print(central_epoch_ra, central_epoch_dec)
    fitter = astro.fitter
    cov_matrix = fitter.evaluate_cov_matrix(central_epoch_ra, central_epoch_dec)
    ra_mura_cov, dec_mudec_cov = cov_matrix[0, 2], cov_matrix[1, 3]
    # do a brute force evaluation of all reasonable central epochs
    epoch_t = np.linspace(1991, 1992, 200)
    ra_vals, dec_vals = [], []
    for t in epoch_t:
        cov = fitter.evaluate_cov_matrix(t, t)
        ra_vals.append(cov[0, 2])
        dec_vals.append(cov[1, 3])
    # assert that the optimal central epochs give better covariances than all of those.
    assert np.all(np.abs(dec_vals) >= dec_mudec_cov)
    assert np.all(np.abs(ra_vals) >= ra_mura_cov)
    assert np.allclose([dec_mudec_cov, ra_mura_cov], 0, atol=1e-9)



"""
Module for generating the chi-squared matrix (and vectors) for the N parameter fit to the epoch astrometry.
"""

import numpy as np
from htof.polynomial import polynomial

FIT_BASIS = polynomial.TaylorSeries
FIT_VANDER = polynomial.taylorvander
# FIT_VANDER must follow np.polynomial.polynomial.Polynomial and
# np.polynomial.polynomial.polyvander syntax, respectively. For example, one could set:
# FIT_BASIS = np.polynomial.polynomial.Polynomial
# FIT_VANDER = np.polynomial.polynomial.polyvander
# if they wanted to use a polynomial basis without the 1/2, 1/6 etc... prefactors.


def _evaluate_basis_functions(w_ra, w_dec, ra_t, dec_t, vander, deg):
    f = np.hstack([[w_ra], np.zeros(2*deg + 2)])
    g = np.hstack([[w_dec], np.zeros(2*deg + 2)])
    f[1:][::2], g[1:][1::2] = vander(ra_t, deg), vander(dec_t, deg)
    return f, g


def ra_sol_vec(a, b, c, d, ra_t, dec_t, w_ra=0, w_dec=0, vander=FIT_VANDER, deg=3):
    """
    :params floats a,b,c,d: components of the inverse covariance matrix for the observation.
            i.e. the ivar matrix should be np.array([[a, b],[c, d]])
    :param ra_t: delta t for right ascension
    :param dec_t: delta t declination
    :param w_ra: pertubation from parallax alone for right ascension
    :param w_dec: pertubation from parallax alone for declination
    :param deg: degree for the fit in ra and dec.
    :param vander: method such that vander(t, degree) returns an array of shape (1, degree+1) with the basis
    functions evaluated at the time t. E.g. if vander= np.polynomial.polynomial.polyvander then basis(5, 3) returns
                  array([[  1.,   5.,  25., 125.]])
    This return is typically called the Vandermonde matrix. E.g. for Legendre polynomial basis we would feed
    vander=np.polynomial.legendre.legvander
    :return:
    """
    f, g = _evaluate_basis_functions(w_ra, w_dec, ra_t, dec_t, vander=vander, deg=deg)
    ra_vec = np.array([a*f[i] + (b+c)/2*g[i] for i in range(2*deg + 3)], dtype=float)
    return ra_vec


def dec_sol_vec(a, b, c, d, ra_t, dec_t, w_ra=0, w_dec=0, vander=FIT_VANDER, deg=3):
    """
    :params floats a,b,c,d: components of the inverse covariance matrix for the observation.
            i.e. the ivar matrix should be np.array([[a, b],[c, d]])
    :param ra_t: delta t for right ascension
    :param dec_t: delta t declination
    :param w_ra: pertubation from parallax alone for right ascension
    :param w_dec: pertubation from parallax alone for declination
    :param deg: degree for the fit in ra and dec.
    :param vander: method such that vander(t, degree) returns an array of shape (1, degree+1) with the basis
    functions evaluated at the time t. E.g. if vander= np.polynomial.polynomial.polyvander then basis(5, 3) returns
                  array([[  1.,   5.,  25., 125.]])
    This return is typically called the Vandermonde matrix. E.g. for Legendre polynomial basis we would feed
    vander=np.polynomial.legendre.legvander
    :return:
    """
    f, g = _evaluate_basis_functions(w_ra, w_dec, ra_t, dec_t, vander=vander, deg=deg)
    dec_vec = np.array([(b+c)/2*f[i] + d*g[i] for i in range(2*deg + 3)], dtype=float)
    return dec_vec


def chi2_matrix(a, b, c, d, ra_t, dec_t, w_ra=0, w_dec=0, vander=FIT_VANDER, deg=3):
    """
    :params floats a,b,c,d: components of the inverse covariance matrix for the observation.
            i.e. the ivar matrix should be np.array([[a, b],[c, d]])
    :param ra_t: delta t for right ascension
    :param dec_t: delta t declination
    :param w_ra: pertubation from parallax alone for right ascension
    :param w_dec: pertubation from parallax alone for declination
    :param deg: degree for the fit in ra and dec.
    :param vander: method such that vander(t, degree) returns an array of shape (1, degree+1) with the basis
    functions evaluated at the time t. E.g. if vander= np.polynomial.polynomial.polyvander then basis(5, 3) returns
                  array([[  1.,   5.,  25., 125.]])
    This return is typically called the Vandermonde matrix. E.g. for Legendre polynomial basis we would feed
    vander=np.polynomial.legendre.legvander
    :return:
    """

    A = np.zeros((2*deg + 3, 2*deg + 3), dtype=np.float32)
    f, g = _evaluate_basis_functions(w_ra, w_dec, ra_t, dec_t, vander=vander, deg=deg)
    # note that b = c for any realistic covariance (or inverse covariance) matrix.
    for k in range(A.shape[0]):
        A[k] = [f[i] * a * f[k] + g[i] * (b+c)/2 * f[k] +
                f[i] * (b+c)/2 * g[k] + g[i] * d * g[k] for i in range(2*deg + 3)]
    return np.array(A, dtype=float)


def transform_coefficients_to_unnormalized_domain(coeffs, ra_min_t, ra_max_t, dec_min_t, dec_max_t,
                                                  use_parallax, old_domain=None, basis=FIT_BASIS):
    # Using basis= taylor does not convert from normalized to unnormalized properly.
    # TODO fix htof.polynomial.polynomial.TaylorSeries so that its .convert() method works properly
    basis = np.polynomial.polynomial.Polynomial
    if old_domain is None:
        old_domain = [-1, 1]
    padded = np.pad(coeffs[1 * use_parallax:], (0, 2 * 3 + 2 - len(coeffs[1 * use_parallax:])),
                    mode='constant', constant_values=0)
    ra_coeffs, dec_coeffs = padded[::2], padded[1::2]
    ra_poly = basis(ra_coeffs, domain=[ra_min_t, ra_max_t], window=[-1, 1])
    dec_poly = basis(dec_coeffs, domain=[dec_min_t, dec_max_t], window=[-1, 1])

    new_coeffs = np.copy(coeffs).astype(np.float64)
    new_coeffs[1 * use_parallax::2] = ra_poly.convert(domain=old_domain).coef
    new_coeffs[1 * use_parallax + 1::2] = dec_poly.convert(domain=old_domain).coef
    # probably need to return the scl of the .mapparms here in order to calculate the unnormed covariance matrix.
    return new_coeffs


def chisq_of_fit(coeffs, ra, dec, ra_epochs, dec_epochs, inv_covs, ra_plx=None,
                 dec_plx=None, use_parallax=True, basis=FIT_BASIS):
    ra_model = basis(coeffs[1 * use_parallax:][::2])(ra_epochs)
    dec_model = basis(coeffs[1 * use_parallax:][1::2])(dec_epochs)
    if use_parallax:
        ra_model += coeffs[0] * ra_plx
        dec_model += coeffs[0] * dec_plx

    modelminusdata = np.hstack([(ra_model - ra).reshape(-1, 1), (dec_model - dec).reshape(-1, 1)])
    chisquared = 0
    for i in range(len(ra_model)):
        chisquared += np.matmul(np.matmul(modelminusdata[i], inv_covs[i]), modelminusdata[i])
    return chisquared


from setuptools import setup, find_packages

setup(name='htof',
      author='G. Mirek Brandt, Daniel Michalik',
      version='0.3.4',
      python_requires='>=3.6',
      packages=find_packages(),
      package_dir={'htof': 'htof'},
      package_data={'htof': ['data/hip1_flagged.txt', 'data/*.csv', 'data/hip2_dvd_flagged.txt']},
      setup_requires=['pytest-runner'],
      install_requires=['astropy>=2.0', 'pandas>=0.24.0', 'scipy>=1.0.0', 'numpy>=1.16'],
      tests_require=['pytest>=3.5'])
